"""Test the sphinx.environment.adapters.indexentries."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.environment.adapters.indexentries import IndexEntries
from sphinx.testing import restructuredtext

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_single_index(app: SphinxTestApp) -> None:
    text = (
        '.. index:: docutils\n'
        '.. index:: Python\n'
        '.. index:: pip; install\n'
        '.. index:: pip; upgrade\n'
        '.. index:: Sphinx\n'
        '.. index:: Ель\n'
        '.. index:: ёлка\n'
        '.. index:: \N{RIGHT-TO-LEFT MARK}\u05e2\u05d1\u05e8\u05d9\u05ea\N{LEFT-TO-RIGHT MARK}\n'
        '.. index:: 9-symbol\n'
        '.. index:: &-symbol\n'
        '.. index:: £100\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 6
    assert index[0] == (
        'Symbols',
        [
            ('&-symbol', ([('', '#index-9')], [], None)),
            ('9-symbol', ([('', '#index-8')], [], None)),
            ('£100', ([('', '#index-10')], [], None)),
        ],
    )
    assert index[1] == ('D', [('docutils', ([('', '#index-0')], [], None))])
    assert index[2] == (
        'P',
        [
            (
                'pip',
                (
                    [],
                    [('install', [('', '#index-2')]), ('upgrade', [('', '#index-3')])],
                    None,
                ),
            ),
            ('Python', ([('', '#index-1')], [], None)),
        ],
    )
    assert index[3] == ('S', [('Sphinx', ([('', '#index-4')], [], None))])
    assert index[4] == (
        'Е',
        [
            ('ёлка', ([('', '#index-6')], [], None)),
            ('Ель', ([('', '#index-5')], [], None)),
        ],
    )
    # Here the word starts with U+200F RIGHT-TO-LEFT MARK, which should be
    # ignored when getting the first letter.
    assert index[5] == (
        '\u05e2',
        [
            (
                '\N{RIGHT-TO-LEFT MARK}\u05e2\u05d1\u05e8\u05d9\u05ea\N{LEFT-TO-RIGHT MARK}',
                ([('', '#index-7')], [], None),
            )
        ],
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_pair_index(app: SphinxTestApp) -> None:
    text = (
        '.. index:: pair: docutils; reStructuredText\n'
        '.. index:: pair: Python; interpreter\n'
        '.. index:: pair: Sphinx; documentation tool\n'
        '.. index:: pair: Sphinx; :+1:\n'
        '.. index:: pair: Sphinx; Ель\n'
        '.. index:: pair: Sphinx; ёлка\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 7
    assert index[0] == (
        'Symbols',
        [(':+1:', ([], [('Sphinx', [('', '#index-3')])], None))],
    )
    assert index[1] == (
        'D',
        [
            ('documentation tool', ([], [('Sphinx', [('', '#index-2')])], None)),
            ('docutils', ([], [('reStructuredText', [('', '#index-0')])], None)),
        ],
    )
    assert index[2] == (
        'I',
        [('interpreter', ([], [('Python', [('', '#index-1')])], None))],
    )
    assert index[3] == (
        'P',
        [('Python', ([], [('interpreter', [('', '#index-1')])], None))],
    )
    assert index[4] == (
        'R',
        [('reStructuredText', ([], [('docutils', [('', '#index-0')])], None))],
    )
    assert index[5] == (
        'S',
        [
            (
                'Sphinx',
                (
                    [],
                    [
                        (':+1:', [('', '#index-3')]),
                        ('documentation tool', [('', '#index-2')]),
                        ('ёлка', [('', '#index-5')]),
                        ('Ель', [('', '#index-4')]),
                    ],
                    None,
                ),
            )
        ],
    )
    assert index[6] == (
        'Е',
        [
            ('ёлка', ([], [('Sphinx', [('', '#index-5')])], None)),
            ('Ель', ([], [('Sphinx', [('', '#index-4')])], None)),
        ],
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_triple_index(app: SphinxTestApp) -> None:
    text = '.. index:: triple: foo; bar; baz\n.. index:: triple: Python; Sphinx; reST\n'
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 5
    assert index[0] == (
        'B',
        [
            ('bar', ([], [('baz, foo', [('', '#index-0')])], None)),
            ('baz', ([], [('foo bar', [('', '#index-0')])], None)),
        ],
    )
    assert index[1] == ('F', [('foo', ([], [('bar baz', [('', '#index-0')])], None))])
    assert index[2] == (
        'P',
        [('Python', ([], [('Sphinx reST', [('', '#index-1')])], None))],
    )
    assert index[3] == (
        'R',
        [('reST', ([], [('Python Sphinx', [('', '#index-1')])], None))],
    )
    assert index[4] == (
        'S',
        [('Sphinx', ([], [('reST, Python', [('', '#index-1')])], None))],
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_see_index(app: SphinxTestApp) -> None:
    text = (
        '.. index:: see: docutils; reStructuredText\n'
        '.. index:: see: Python; interpreter\n'
        '.. index:: see: Sphinx; documentation tool\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 3
    assert index[0] == ('D', [('docutils', ([], [('see reStructuredText', [])], None))])
    assert index[1] == ('P', [('Python', ([], [('see interpreter', [])], None))])
    assert index[2] == ('S', [('Sphinx', ([], [('see documentation tool', [])], None))])


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_seealso_index(app: SphinxTestApp) -> None:
    text = (
        '.. index:: seealso: docutils; reStructuredText\n'
        '.. index:: seealso: Python; interpreter\n'
        '.. index:: seealso: Sphinx; documentation tool\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 3
    assert index[0] == (
        'D',
        [('docutils', ([], [('see also reStructuredText', [])], None))],
    )
    assert index[1] == ('P', [('Python', ([], [('see also interpreter', [])], None))])
    assert index[2] == (
        'S',
        [('Sphinx', ([], [('see also documentation tool', [])], None))],
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_main_index(app: SphinxTestApp) -> None:
    text = (
        '.. index:: !docutils\n'
        '.. index:: docutils\n'
        '.. index:: pip; install\n'
        '.. index:: !pip; install\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 2
    assert index[0] == (
        'D',
        [('docutils', ([('main', '#index-0'), ('', '#index-1')], [], None))],
    )
    assert index[1] == (
        'P',
        [('pip', ([], [('install', [('main', '#index-3'), ('', '#index-2')])], None))],
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_index_with_name(app: SphinxTestApp) -> None:
    text = (
        '.. index:: single: docutils\n'
        '   :name: ref1\n'
        '.. index:: single: Python\n'
        '   :name: ref2\n'
        '.. index:: Sphinx\n'
    )
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)

    # check index is created correctly
    assert len(index) == 3
    assert index[0] == ('D', [('docutils', ([('', '#ref1')], [], None))])
    assert index[1] == ('P', [('Python', ([('', '#ref2')], [], None))])
    assert index[2] == ('S', [('Sphinx', ([('', '#index-0')], [], None))])

    # check the reference labels are created correctly
    std = app.env.domains.standard_domain
    assert std.anonlabels['ref1'] == ('index', 'ref1')
    assert std.anonlabels['ref2'] == ('index', 'ref2')


@pytest.mark.sphinx(
    'dummy',
    testroot='root',
    freshenv=True,
)
def test_create_index_by_key(app: SphinxTestApp) -> None:
    # At present, only glossary directive is able to create index key
    text = '.. glossary::\n\n   docutils\n   Python\n   スフィンクス : ス\n'
    restructuredtext.parse(app, text)
    index = IndexEntries(app.env).create_index(app.builder)
    assert len(index) == 3
    assert index[0] == ('D', [('docutils', ([('main', '#term-docutils')], [], None))])
    assert index[1] == ('P', [('Python', ([('main', '#term-Python')], [], None))])
    assert index[2] == ('ス', [('スフィンクス', ([('main', '#term-0')], [], 'ス'))])



"""Test all builders."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx.errors import SphinxError

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


def test_root_doc_not_found(tmp_path, make_app):
    (tmp_path / 'conf.py').touch()
    assert [p.name for p in tmp_path.iterdir()] == ['conf.py']

    app = make_app('dummy', srcdir=tmp_path)
    with pytest.raises(SphinxError):
        app.build(force_all=True)  # no index.rst


@pytest.mark.sphinx('text', testroot='circular')
def test_circular_toctree(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    warnings = app.warning.getvalue()
    assert (
        'circular toctree references detected, ignoring: sub <- index <- sub'
    ) in warnings
    assert (
        'circular toctree references detected, ignoring: index <- sub <- index'
    ) in warnings


@pytest.mark.sphinx('text', testroot='numbered-circular')
def test_numbered_circular_toctree(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    warnings = app.warning.getvalue()
    assert (
        'circular toctree references detected, ignoring: sub <- index <- sub'
    ) in warnings
    assert (
        'circular toctree references detected, ignoring: index <- sub <- index'
    ) in warnings


@pytest.mark.sphinx('text', testroot='toctree-multiple-parents')
def test_multiple_parents_toctree(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (
        "document is referenced in multiple toctrees: ['bravo', 'delta'], selecting: delta <- charlie"
    ) in app.status.getvalue()


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx('dummy', testroot='images')
def test_image_glob(app):
    app.build(force_all=True)

    # index.rst
    doctree = app.env.get_doctree('index')

    assert isinstance(doctree[0][1], nodes.image)
    assert doctree[0][1]['candidates'] == {'*': 'rimg.png'}
    assert doctree[0][1]['uri'] == 'rimg.png'

    assert isinstance(doctree[0][2], nodes.figure)
    assert isinstance(doctree[0][2][0], nodes.image)
    assert doctree[0][2][0]['candidates'] == {'*': 'rimg.png'}
    assert doctree[0][2][0]['uri'] == 'rimg.png'

    assert isinstance(doctree[0][3], nodes.image)
    assert doctree[0][3]['candidates'] == {
        'application/pdf': 'img.pdf',
        'image/gif': 'img.gif',
        'image/png': 'img.png',
    }
    assert doctree[0][3]['uri'] == 'img.*'

    assert isinstance(doctree[0][4], nodes.figure)
    assert isinstance(doctree[0][4][0], nodes.image)
    assert doctree[0][4][0]['candidates'] == {
        'application/pdf': 'img.pdf',
        'image/gif': 'img.gif',
        'image/png': 'img.png',
    }
    assert doctree[0][4][0]['uri'] == 'img.*'

    # subdir/index.rst
    doctree = app.env.get_doctree('subdir/index')

    assert isinstance(doctree[0][1], nodes.image)
    assert doctree[0][1]['candidates'] == {'*': 'subdir/rimg.png'}
    assert doctree[0][1]['uri'] == 'subdir/rimg.png'

    assert isinstance(doctree[0][2], nodes.image)
    assert doctree[0][2]['candidates'] == {
        'application/pdf': 'subdir/svgimg.pdf',
        'image/svg+xml': 'subdir/svgimg.svg',
    }
    assert doctree[0][2]['uri'] == 'subdir/svgimg.*'

    assert isinstance(doctree[0][3], nodes.figure)
    assert isinstance(doctree[0][3][0], nodes.image)
    assert doctree[0][3][0]['candidates'] == {
        'application/pdf': 'subdir/svgimg.pdf',
        'image/svg+xml': 'subdir/svgimg.svg',
    }
    assert doctree[0][3][0]['uri'] == 'subdir/svgimg.*'



from __future__ import annotations

from sphinx.errors import ExtensionError


def test_extension_error_repr() -> None:
    exc = ExtensionError('foo')
    assert repr(exc) == "ExtensionError('foo')"


def test_extension_error_with_orig_exc_repr() -> None:
    exc = ExtensionError('foo', Exception('bar'))
    assert repr(exc) == "ExtensionError('foo', Exception('bar'))"



"""Test the ``UnreferencedFootnotesDetector`` transform."""

from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from pathlib import Path

    from sphinx.testing.util import SphinxTestApp


def test_warnings(make_app: type[SphinxTestApp], tmp_path: Path) -> None:
    """Test that warnings are emitted for unreferenced footnotes."""
    tmp_path.joinpath('conf.py').touch()
    tmp_path.joinpath('index.rst').write_text(
        """
Title
=====
[1]_ [#label2]_

.. [1] This is a normal footnote.
.. [2] This is a normal footnote.
.. [2] This is a normal footnote.
.. [3] This is a normal footnote.
.. [*] This is a symbol footnote.
.. [#] This is an auto-numbered footnote.
.. [#label1] This is an auto-numbered footnote with a label.
.. [#label1] This is an auto-numbered footnote with a label.
.. [#label2] This is an auto-numbered footnote with a label.
        """,
        encoding='utf8',
    )
    app = make_app(srcdir=tmp_path)
    app.build()
    warnings = strip_escape_sequences(app.warning.getvalue()).lstrip()
    warnings = warnings.replace(str(tmp_path / 'index.rst'), 'source/index.rst')
    assert (
        warnings
        == """\
source/index.rst:8: WARNING: Duplicate explicit target name: "2". [docutils]
source/index.rst:13: WARNING: Duplicate explicit target name: "label1". [docutils]
source/index.rst:9: WARNING: Footnote [3] is not referenced. [ref.footnote]
source/index.rst:10: WARNING: Footnote [*] is not referenced. [ref.footnote]
source/index.rst:11: WARNING: Footnote [#] is not referenced. [ref.footnote]
"""
    )



"""Test sphinx.ext.viewcode extension."""

from __future__ import annotations

import re
import shutil
from typing import TYPE_CHECKING

import pygments
import pytest

from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


def check_viewcode_output(app: SphinxTestApp) -> str:
    if tuple(map(int, pygments.__version__.split('.')[:2])) >= (2, 19):
        sp = '<span> </span>'
    else:
        sp = ' '

    warnings = re.sub(r'\\+', '/', app.warning.getvalue())
    assert re.findall(
        r"index.rst:\d+: WARNING: Object named 'func1' not found in include "
        r"file .*/spam/__init__.py'",
        warnings,
    )

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert result.count('href="_modules/spam/mod1.html#func1"') == 2
    assert result.count('href="_modules/spam/mod2.html#func2"') == 2
    assert result.count('href="_modules/spam/mod1.html#Class1"') == 2
    assert result.count('href="_modules/spam/mod2.html#Class2"') == 2
    assert result.count('@decorator') == 1

    # test that the class attribute is correctly documented
    assert result.count('this is Class3') == 2
    assert 'this is the class attribute class_attr' in result
    # the next assert fails, until the autodoc bug gets fixed
    assert result.count('this is the class attribute class_attr') == 2

    result = (app.outdir / '_modules/spam/mod1.html').read_text(encoding='utf8')
    # filter pygments classes
    result = re.sub('<span class="[^"]{,2}">', '<span>', result)
    assert (
        '<div class="viewcode-block" id="Class1">\n'
        '<a class="viewcode-back" href="../../index.html#spam.Class1">[docs]</a>\n'
    ) in result
    assert '<span>@decorator</span>\n' in result
    assert f'<span>class</span>{sp}<span>Class1</span><span>:</span>\n' in result
    assert (
        '<span>    </span>'
        '<span>&quot;&quot;&quot;this is Class1&quot;&quot;&quot;</span></div>\n'
    ) in result

    return result


@pytest.mark.sphinx(
    'html',
    testroot='ext-viewcode',
    freshenv=True,
    confoverrides={'viewcode_line_numbers': True},
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_viewcode_linenos(app: SphinxTestApp) -> None:
    shutil.rmtree(app.outdir / '_modules', ignore_errors=True)
    app.build(force_all=True)

    result = check_viewcode_output(app)
    assert '<span class="linenos"> 1</span>' in result


@pytest.mark.sphinx(
    'html',
    testroot='ext-viewcode',
    freshenv=True,
    confoverrides={'viewcode_line_numbers': False},
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_viewcode(app: SphinxTestApp) -> None:
    shutil.rmtree(app.outdir / '_modules', ignore_errors=True)
    app.build(force_all=True)

    result = check_viewcode_output(app)
    assert 'class="linenos">' not in result


@pytest.mark.sphinx('epub', testroot='ext-viewcode')
@pytest.mark.usefixtures('rollback_sysmodules')
def test_viewcode_epub_default(app: SphinxTestApp) -> None:
    shutil.rmtree(app.outdir)
    app.build(force_all=True)

    assert not (app.outdir / '_modules/spam/mod1.xhtml').exists()

    result = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert result.count('href="_modules/spam/mod1.xhtml#func1"') == 0


@pytest.mark.sphinx(
    'epub',
    testroot='ext-viewcode',
    confoverrides={'viewcode_enable_epub': True},
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_viewcode_epub_enabled(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    assert (app.outdir / '_modules/spam/mod1.xhtml').exists()

    result = (app.outdir / 'index.xhtml').read_text(encoding='utf8')
    assert result.count('href="_modules/spam/mod1.xhtml#func1"') == 2


@pytest.mark.sphinx('html', testroot='ext-viewcode', tags=['test_linkcode'])
def test_linkcode(app: SphinxTestApp) -> None:
    app.build(filenames=[app.srcdir / 'objects.rst'])

    stuff = (app.outdir / 'objects.html').read_text(encoding='utf8')

    assert 'https://foobar/source/foolib.py' in stuff
    assert 'https://foobar/js/' in stuff
    assert 'https://foobar/c/' in stuff
    assert 'https://foobar/cpp/' in stuff
    assert 'http://foobar/rst/' in stuff


@pytest.mark.sphinx('html', testroot='ext-viewcode-find', freshenv=True)
def test_local_source_files(app: SphinxTestApp) -> None:
    def find_source(app, modname):
        if modname == 'not_a_package':
            source = app.srcdir / 'not_a_package/__init__.py'
            tags = {
                'func1': ('def', 1, 1),
                'Class1': ('class', 1, 1),
                'not_a_package.submodule.func1': ('def', 1, 1),
                'not_a_package.submodule.Class1': ('class', 1, 1),
            }
        else:
            source = app.srcdir / 'not_a_package/submodule.py'
            tags = {
                'not_a_package.submodule.func1': ('def', 11, 15),
                'Class1': ('class', 19, 22),
                'not_a_package.submodule.Class1': ('class', 19, 22),
                'Class3': ('class', 25, 30),
                'not_a_package.submodule.Class3.class_attr': ('other', 29, 29),
            }
        return source.read_text(encoding='utf8'), tags

    app.connect('viewcode-find-source', find_source)
    app.build(force_all=True)

    warnings = re.sub(r'\\+', '/', app.warning.getvalue())
    assert re.findall(
        r"index.rst:\d+: WARNING: Object named 'func1' not found in include "
        r"file .*/not_a_package/__init__.py'",
        warnings,
    )

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    for needle in (
        'href="_modules/not_a_package.html#func1"',
        'href="_modules/not_a_package.html#not_a_package.submodule.func1"',
        'href="_modules/not_a_package/submodule.html#Class1"',
        'href="_modules/not_a_package/submodule.html#Class3"',
        'href="_modules/not_a_package/submodule.html#not_a_package.submodule.Class1"',
        'href="_modules/not_a_package/submodule.html#not_a_package.submodule.Class3.class_attr"',
        'This is the class attribute class_attr',
    ):
        assert result.count(needle) == 1


@pytest.mark.sphinx('html', testroot='ext-viewcode-find-package', freshenv=True)
def test_find_local_package_import_path(app, status, warning):
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')

    count_func1 = result.count(
        'href="_modules/main_package/subpackage/_subpackage2/submodule.html#func1"'
    )
    assert count_func1 == 1

    count_class1 = result.count(
        'href="_modules/main_package/subpackage/_subpackage2/submodule.html#Class1"'
    )
    assert count_class1 == 1

    count_class3 = result.count(
        'href="_modules/main_package/subpackage/_subpackage2/submodule.html#Class3"'
    )
    assert count_class3 == 1



"""Test the BuildEnvironment class."""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.builders.latex import LaTeXBuilder
from sphinx.config import Config
from sphinx.environment import (
    CONFIG_CHANGED,
    CONFIG_EXTENSIONS_CHANGED,
    CONFIG_NEW,
    CONFIG_OK,
    _differing_config_keys,
)

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('dummy', testroot='basic', copy_test_root=True)
def test_config_status(make_app, app_params):
    args, kwargs = app_params

    # clean build
    app1 = make_app(*args, freshenv=True, **kwargs)
    assert app1.env.config_status == CONFIG_NEW
    app1.build()
    output = strip_escape_sequences(app1.status.getvalue())
    # assert 'The configuration has changed' not in output
    assert '[new config] 1 added' in output

    # incremental build (no config changed)
    app2 = make_app(*args, **kwargs)
    assert app2.env.config_status == CONFIG_OK
    app2.build()
    output = strip_escape_sequences(app2.status.getvalue())
    assert 'The configuration has changed' not in output
    assert '0 added, 0 changed, 0 removed' in output

    # incremental build (config entry changed)
    app3 = make_app(*args, confoverrides={'root_doc': 'indexx'}, **kwargs)
    fname = app3.srcdir / 'index.rst'
    other_fname = app3.srcdir / 'indexx.rst'
    assert fname.is_file()
    shutil.move(fname, other_fname)
    assert app3.env.config_status == CONFIG_CHANGED
    app3.build()
    shutil.move(other_fname, fname)
    output = strip_escape_sequences(app3.status.getvalue())
    assert 'The configuration has changed' in output
    assert "[config changed ('master_doc')] 1 added," in output

    # incremental build (extension changed)
    app4 = make_app(
        *args, confoverrides={'extensions': ['sphinx.ext.autodoc']}, **kwargs
    )
    assert app4.env.config_status == CONFIG_EXTENSIONS_CHANGED
    app4.build()
    want_str = "[extensions changed ('sphinx.ext.autodoc')] 1 added"
    output = strip_escape_sequences(app4.status.getvalue())
    assert 'The configuration has changed' not in output
    assert want_str in output


@pytest.mark.sphinx('dummy', testroot='root')
def test_images(app: SphinxTestApp) -> None:
    app.build()

    tree = app.env.get_doctree('images')
    htmlbuilder = StandaloneHTMLBuilder(app, app.env)
    htmlbuilder.init()
    htmlbuilder.imgpath = 'dummy'
    htmlbuilder.post_process_images(tree)
    assert set(htmlbuilder.images.keys()) == {
        'subdir/img.png',
        'img.png',
        'subdir/simg.png',
        'svgimg.svg',
        'img.foo.png',
    }
    assert set(htmlbuilder.images.values()) == {
        'img.png',
        'img1.png',
        'simg.png',
        'svgimg.svg',
        'img.foo.png',
    }

    latexbuilder = LaTeXBuilder(app, app.env)
    latexbuilder.init()
    latexbuilder.post_process_images(tree)
    assert set(latexbuilder.images.keys()) == {
        'subdir/img.png',
        'subdir/simg.png',
        'img.png',
        'img.pdf',
        'svgimg.pdf',
        'img.foo.png',
    }
    assert set(latexbuilder.images.values()) == {
        'img.pdf',
        'img.png',
        'img1.png',
        'simg.png',
        'svgimg.pdf',
        'img.foo.png',
    }


@pytest.mark.sphinx('dummy', testroot='root')
def test_object_inventory(app: SphinxTestApp) -> None:
    app.build()
    refs = app.env.domaindata['py']['objects']

    assert 'func_without_module' in refs
    assert refs['func_without_module'] == (
        'objects',
        'func_without_module',
        'function',
        False,
    )
    assert 'func_without_module2' in refs
    assert 'mod.func_in_module' in refs
    assert 'mod.Cls' in refs
    assert 'mod.Cls.meth1' in refs
    assert 'mod.Cls.meth2' in refs
    assert 'mod.Cls.meths' in refs

    assert 'mod.Error' not in refs
    assert 'errmod.Error' in refs

    assert 'func_in_module' not in refs
    assert 'func_noindex' not in refs

    assert app.env.domaindata['py']['modules']['mod'] == (
        'objects',
        'module-mod',
        'Module synopsis.',
        'UNIX',
        False,
    )

    assert app.env.domains.python_domain.data is app.env.domaindata['py']
    assert app.env.domains.c_domain.data is app.env.domaindata['c']


@pytest.mark.sphinx('dummy', testroot='basic')
def test_env_relfn2path(app: SphinxTestApp) -> None:
    # relative filename and root document
    relfn, absfn = app.env.relfn2path('logo.jpg', 'index')
    assert relfn == 'logo.jpg'
    assert absfn == str(app.srcdir / 'logo.jpg')

    # absolute filename and root document
    relfn, absfn = app.env.relfn2path('/logo.jpg', 'index')
    assert relfn == 'logo.jpg'
    assert absfn == str(app.srcdir / 'logo.jpg')

    # relative filename and a document in subdir
    relfn, absfn = app.env.relfn2path('logo.jpg', 'subdir/index')
    assert Path(relfn) == Path('subdir/logo.jpg')
    assert absfn == str(app.srcdir / 'subdir' / 'logo.jpg')

    # absolute filename and a document in subdir
    relfn, absfn = app.env.relfn2path('/logo.jpg', 'subdir/index')
    assert relfn == 'logo.jpg'
    assert absfn == str(app.srcdir / 'logo.jpg')

    # relative filename having subdir
    relfn, absfn = app.env.relfn2path('images/logo.jpg', 'index')
    assert relfn == 'images/logo.jpg'
    assert absfn == str(app.srcdir / 'images' / 'logo.jpg')

    # relative path traversal
    relfn, absfn = app.env.relfn2path('../logo.jpg', 'index')
    assert relfn == '../logo.jpg'
    assert absfn == str(app.srcdir.parent / 'logo.jpg')

    # relative path traversal
    relfn, absfn = app.env.relfn2path('subdir/../logo.jpg', 'index')
    assert relfn == 'logo.jpg'
    assert absfn == str(app.srcdir / 'logo.jpg')

    # omit docname (w/ current docname)
    app.env.current_document.docname = 'subdir/document'
    relfn, absfn = app.env.relfn2path('images/logo.jpg')
    assert Path(relfn) == Path('subdir/images/logo.jpg')
    assert absfn == str(app.srcdir / 'subdir' / 'images' / 'logo.jpg')

    # omit docname (w/o current docname)
    app.env.current_document.clear()
    with pytest.raises(KeyError, match=r"^'docname'$"):
        app.env.relfn2path('images/logo.jpg')


def test_differing_config_keys() -> None:
    diff = _differing_config_keys

    old = Config({'project': 'old'})
    new = Config({'project': 'new'})
    assert diff(old, new) == frozenset({'project'})

    old = Config({'project': 'project', 'release': 'release'})
    new = Config({'project': 'project', 'version': 'version'})
    assert diff(old, new) == frozenset({'release', 'version'})

    old = Config({'project': 'project', 'release': 'release'})
    new = Config({'project': 'project'})
    assert diff(old, new) == frozenset({'release'})

    old = Config({'project': 'project'})
    new = Config({'project': 'project', 'version': 'version'})
    assert diff(old, new) == frozenset({'version'})

    old = Config({'project': 'project', 'release': 'release', 'version': 'version'})
    new = Config({'project': 'project', 'release': 'release', 'version': 'version'})
    assert diff(old, new) == frozenset()

    old = Config({'project': 'old', 'release': 'release'})
    new = Config({'project': 'new', 'version': 'version'})
    assert diff(old, new) == frozenset({'project', 'release', 'version'})



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from xml.etree.ElementTree import Element


def _intradocument_hyperlink_check(nodes: Sequence[Element]) -> None:
    """Confirm that a series of nodes are all HTML hyperlinks to the current page"""
    assert nodes, 'Expected at least one node to check'
    for node in nodes:
        assert node.tag == 'a', 'Attempted to check hyperlink on a non-anchor element'
        href = node.attrib.get('href')
        # Allow Sphinx index and table hyperlinks to be non-same-document, as exceptions.
        if href in {'genindex.html', 'py-modindex.html', 'search.html'}:
            continue
        assert not href or href.startswith('#'), 'Hyperlink failed same-document check'



"""Tests util.inspect functions."""

from __future__ import annotations

import ast
import datetime
import enum
import functools
import types
from inspect import Parameter
from typing import Callable, List, Optional, Union  # NoQA: UP035

import pytest

from sphinx.util import inspect
from sphinx.util.inspect import (
    TypeAliasForwardRef,
    TypeAliasNamespace,
    stringify_signature,
)
from sphinx.util.typing import stringify_annotation


class Base:
    def meth(self):
        pass

    @staticmethod
    def staticmeth():
        pass

    @classmethod
    def classmeth(cls):
        pass

    @property
    def prop(self):
        pass

    partialmeth = functools.partialmethod(meth)

    async def coroutinemeth(self):
        pass

    partial_coroutinemeth = functools.partialmethod(coroutinemeth)

    @classmethod
    async def coroutineclassmeth(cls):
        """A documented coroutine classmethod"""
        pass


class Inherited(Base):
    pass


class MyInt(int):
    @classmethod
    def classmeth(cls):
        pass


class MyIntOverride(MyInt):
    @classmethod
    def from_bytes(cls, *a, **kw):
        return super().from_bytes(*a, **kw)

    def conjugate(self):
        return super().conjugate()


def func():
    pass


async def coroutinefunc():
    pass


async def asyncgenerator():  # NoQA: RUF029
    yield


partial_func = functools.partial(func)
partial_coroutinefunc = functools.partial(coroutinefunc)

builtin_func = print
partial_builtin_func = functools.partial(print)


class Descriptor:
    def __get__(self, obj, typ=None):
        pass


class _Callable:
    def __call__(self):
        pass


def _decorator(f):
    @functools.wraps(f)
    def wrapper():
        return f()

    return wrapper


def test_TypeAliasForwardRef():
    alias = TypeAliasForwardRef('example')
    sig_str = stringify_annotation(alias, 'fully-qualified-except-typing')
    assert sig_str == "TypeAliasForwardRef('example')"

    alias = Optional[alias]  # NoQA: UP045
    sig_str = stringify_annotation(alias, 'fully-qualified-except-typing')
    assert sig_str == "TypeAliasForwardRef('example') | None"


def test_TypeAliasNamespace() -> None:
    import logging.config

    type_alias = TypeAliasNamespace({
        'logging.Filter': 'MyFilter',
        'logging.Handler': 'MyHandler',
        'logging.handlers.SyslogHandler': 'MySyslogHandler',
    })

    assert type_alias['logging'].Filter == 'MyFilter'
    assert type_alias['logging'].Handler == 'MyHandler'
    assert type_alias['logging'].handlers.SyslogHandler == 'MySyslogHandler'
    assert type_alias['logging'].Logger == logging.Logger
    assert type_alias['logging'].config == logging.config

    with pytest.raises(KeyError):
        assert type_alias['log']

    with pytest.raises(KeyError):
        assert type_alias['unknown']


def test_signature() -> None:
    # literals
    with pytest.raises(TypeError):
        inspect.signature(1)  # type: ignore[arg-type]

    with pytest.raises(TypeError):
        inspect.signature('')  # type: ignore[arg-type]

    # builtins are supported on a case-by-case basis, depending on whether
    # they define __text_signature__
    if getattr(list, '__text_signature__', None):
        sig = inspect.stringify_signature(inspect.signature(list))
        assert sig == '(iterable=(), /)'
    else:
        with pytest.raises(ValueError, match='no signature found for builtin type'):
            inspect.signature(list)
    with pytest.raises(ValueError, match='no signature found for builtin type'):
        inspect.signature(range)

    # normal function
    def func(a, b, c=1, d=2, *e, **f):
        pass

    sig = inspect.stringify_signature(inspect.signature(func))
    assert sig == '(a, b, c=1, d=2, *e, **f)'


def test_signature_partial() -> None:
    def fun(a, b, c=1, d=2):
        pass

    p = functools.partial(fun, 10, c=11)

    sig = inspect.signature(p)
    assert stringify_signature(sig) == '(b, *, c=11, d=2)'


def test_signature_methods() -> None:
    class Foo:
        def meth1(self, arg1, **kwargs):
            pass

        @classmethod
        def meth2(cls, arg1, *args, **kwargs):
            pass

        @staticmethod
        def meth3(arg1, *args, **kwargs):
            pass

    @functools.wraps(Foo().meth1)
    def wrapped_bound_method(*args, **kwargs):
        pass

    # unbound method
    sig = inspect.signature(Foo.meth1)
    assert stringify_signature(sig) == '(self, arg1, **kwargs)'

    sig = inspect.signature(Foo.meth1, bound_method=True)
    assert stringify_signature(sig) == '(arg1, **kwargs)'

    # bound method
    sig = inspect.signature(Foo().meth1)
    assert stringify_signature(sig) == '(arg1, **kwargs)'

    # class method
    sig = inspect.signature(Foo.meth2)
    assert stringify_signature(sig) == '(arg1, *args, **kwargs)'

    sig = inspect.signature(Foo().meth2)
    assert stringify_signature(sig) == '(arg1, *args, **kwargs)'

    # static method
    sig = inspect.signature(Foo.meth3)
    assert stringify_signature(sig) == '(arg1, *args, **kwargs)'

    sig = inspect.signature(Foo().meth3)
    assert stringify_signature(sig) == '(arg1, *args, **kwargs)'

    # wrapped bound method
    sig = inspect.signature(wrapped_bound_method)
    assert stringify_signature(sig) == '(arg1, **kwargs)'


def test_signature_partialmethod() -> None:
    from functools import partialmethod

    class Foo:
        def meth1(self, arg1, arg2, arg3=None, arg4=None):
            pass

        def meth2(self, arg1, arg2):
            pass

        foo = partialmethod(meth1, 1, 2)
        bar = partialmethod(meth1, 1, arg3=3)
        baz = partialmethod(meth2, 1, 2)

    subject = Foo()
    sig = inspect.signature(subject.foo)
    assert stringify_signature(sig) == '(arg3=None, arg4=None)'

    sig = inspect.signature(subject.bar)
    assert stringify_signature(sig) == '(arg2, *, arg3=3, arg4=None)'

    sig = inspect.signature(subject.baz)
    assert stringify_signature(sig) == '()'


def test_signature_annotations() -> None:
    import tests.test_util.typing_test_data as mod

    # Class annotations
    sig = inspect.signature(mod.f0)
    assert stringify_signature(sig) == '(x: int, y: numbers.Integral) -> None'

    # Generic types with concrete parameters
    sig = inspect.signature(mod.f1)
    assert stringify_signature(sig) == '(x: list[int]) -> typing.List[int]'

    # TypeVars and generic types with TypeVars
    sig = inspect.signature(mod.f2)
    assert stringify_signature(sig) == (
        '(x: typing.List[tests.test_util.typing_test_data.T],'
        ' y: typing.List[tests.test_util.typing_test_data.T_co],'
        ' z: tests.test_util.typing_test_data.T'
        ') -> typing.List[tests.test_util.typing_test_data.T_contra]'
    )

    # Union types
    sig = inspect.signature(mod.f3)
    assert stringify_signature(sig) == '(x: str | numbers.Integral) -> None'

    # Quoted annotations
    sig = inspect.signature(mod.f4)
    assert stringify_signature(sig) == '(x: str, y: str) -> None'

    # Keyword-only arguments
    sig = inspect.signature(mod.f5)
    assert stringify_signature(sig) == '(x: int, *, y: str, z: str) -> None'

    # Keyword-only arguments with varargs
    sig = inspect.signature(mod.f6)
    assert stringify_signature(sig) == '(x: int, *args, y: str, z: str) -> None'

    # Space around '=' for defaults
    sig = inspect.signature(mod.f7)
    sig_str = stringify_signature(sig)
    assert sig_str == '(x: int = None, y: dict = {}) -> None'

    # Callable types
    sig = inspect.signature(mod.f8)
    assert stringify_signature(sig) == '(x: typing.Callable[[int, str], int]) -> None'

    sig = inspect.signature(mod.f9)
    assert stringify_signature(sig) == '(x: typing.Callable) -> None'

    # Tuple types
    sig = inspect.signature(mod.f10)
    sig_str = stringify_signature(sig)
    assert sig_str == '(x: typing.Tuple[int, str], y: typing.Tuple[int, ...]) -> None'

    # Instance annotations
    sig = inspect.signature(mod.f11)
    assert stringify_signature(sig) == '(x: CustomAnnotation, y: 123) -> None'

    # tuple with more than two items
    sig = inspect.signature(mod.f12)
    assert stringify_signature(sig) == '() -> typing.Tuple[int, str, int]'

    # optional
    sig = inspect.signature(mod.f13)
    assert stringify_signature(sig) == '() -> str | None'

    # optional union
    sig = inspect.signature(mod.f20)
    assert stringify_signature(sig) in {
        '() -> int | str | None',
        '() -> str | int | None',
    }

    # Any
    sig = inspect.signature(mod.f14)
    assert stringify_signature(sig) == '() -> typing.Any'

    # ForwardRef
    sig = inspect.signature(mod.f15)
    assert stringify_signature(sig) == '(x: Unknown, y: int) -> typing.Any'

    # keyword only arguments (1)
    sig = inspect.signature(mod.f16)
    assert stringify_signature(sig) == '(arg1, arg2, *, arg3=None, arg4=None)'

    # keyword only arguments (2)
    sig = inspect.signature(mod.f17)
    assert stringify_signature(sig) == '(*, arg3, arg4)'

    sig = inspect.signature(mod.f18)
    assert stringify_signature(sig) == (
        '(self, arg1: int | typing.Tuple = 10) -> typing.List[typing.Dict]'
    )

    # annotations for variadic and keyword parameters
    sig = inspect.signature(mod.f19)
    assert stringify_signature(sig) == '(*args: int, **kwargs: str)'

    # default value is inspect.Signature.empty
    sig = inspect.signature(mod.f21)
    assert stringify_signature(sig) == "(arg1='whatever', arg2)"

    # type hints by string
    sig = inspect.signature(mod.Node.children)
    sig_str = stringify_signature(sig)
    assert sig_str == '(self) -> typing.List[tests.test_util.typing_test_data.Node]'

    sig = inspect.signature(mod.Node.__init__)
    sig_str = stringify_signature(sig)
    assert sig_str == (
        '(self, parent: tests.test_util.typing_test_data.Node | None) -> None'
    )

    # show_annotation is False
    sig = inspect.signature(mod.f7)
    assert stringify_signature(sig, show_annotation=False) == '(x=None, y={})'

    # show_return_annotation is False
    sig = inspect.signature(mod.f7)
    sig_str = stringify_signature(sig, show_return_annotation=False)
    assert sig_str == '(x: int = None, y: dict = {})'

    # unqualified_typehints is True
    sig = inspect.signature(mod.f7)
    sig_str = stringify_signature(sig, unqualified_typehints=True)
    assert sig_str == '(x: int = None, y: dict = {}) -> None'

    # case: separator at head
    sig = inspect.signature(mod.f22)
    assert stringify_signature(sig) == '(*, a, b)'

    # case: separator in the middle
    sig = inspect.signature(mod.f23)
    assert stringify_signature(sig) == '(a, b, /, c, d)'

    sig = inspect.signature(mod.f24)
    assert stringify_signature(sig) == '(a, /, *, b)'

    # case: separator at tail
    sig = inspect.signature(mod.f25)
    assert stringify_signature(sig) == '(a, b, /)'

    # collapse Literal types
    sig = inspect.signature(mod.f26)
    sig_str = stringify_signature(sig)
    assert sig_str == (
        "(x: typing.Literal[1, 2, 3] = 1, y: typing.Literal['a', 'b'] = 'a') -> None"
    )


def test_signature_from_str_basic() -> None:
    signature = '(a, b, *args, c=0, d="blah", **kwargs)'
    sig = inspect.signature_from_str(signature)
    assert list(sig.parameters.keys()) == ['a', 'b', 'args', 'c', 'd', 'kwargs']
    assert sig.parameters['a'].name == 'a'
    assert sig.parameters['a'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['a'].default == Parameter.empty
    assert sig.parameters['a'].annotation == Parameter.empty
    assert sig.parameters['b'].name == 'b'
    assert sig.parameters['b'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['b'].default == Parameter.empty
    assert sig.parameters['b'].annotation == Parameter.empty
    assert sig.parameters['args'].name == 'args'
    assert sig.parameters['args'].kind == Parameter.VAR_POSITIONAL
    assert sig.parameters['args'].default == Parameter.empty
    assert sig.parameters['args'].annotation == Parameter.empty
    assert sig.parameters['c'].name == 'c'
    assert sig.parameters['c'].kind == Parameter.KEYWORD_ONLY
    assert sig.parameters['c'].default == '0'
    assert sig.parameters['c'].annotation == Parameter.empty
    assert sig.parameters['d'].name == 'd'
    assert sig.parameters['d'].kind == Parameter.KEYWORD_ONLY
    assert sig.parameters['d'].default == "'blah'"
    assert sig.parameters['d'].annotation == Parameter.empty
    assert sig.parameters['kwargs'].name == 'kwargs'
    assert sig.parameters['kwargs'].kind == Parameter.VAR_KEYWORD
    assert sig.parameters['kwargs'].default == Parameter.empty
    assert sig.parameters['kwargs'].annotation == Parameter.empty
    assert sig.return_annotation == Parameter.empty


def test_signature_from_str_default_values() -> None:
    signature = (
        '(a=0, b=0.0, c="str", d=b"bytes", e=..., f=True, '
        'g=[1, 2, 3], h={"a": 1}, i={1, 2, 3}, '
        'j=lambda x, y: None, k=None, l=object(), m=foo.bar.CONSTANT)'
    )
    sig = inspect.signature_from_str(signature)
    assert sig.parameters['a'].default == '0'
    assert sig.parameters['b'].default == '0.0'
    assert sig.parameters['c'].default == "'str'"
    assert sig.parameters['d'].default == "b'bytes'"
    assert sig.parameters['e'].default == '...'
    assert sig.parameters['f'].default == 'True'
    assert sig.parameters['g'].default == '[1, 2, 3]'
    assert sig.parameters['h'].default == "{'a': 1}"
    assert sig.parameters['i'].default == '{1, 2, 3}'
    assert sig.parameters['j'].default == 'lambda x, y: ...'
    assert sig.parameters['k'].default == 'None'
    assert sig.parameters['l'].default == 'object()'
    assert sig.parameters['m'].default == 'foo.bar.CONSTANT'


def test_signature_from_str_annotations() -> None:
    signature = '(a: int, *args: bytes, b: str = "blah", **kwargs: float) -> None'
    sig = inspect.signature_from_str(signature)
    assert list(sig.parameters.keys()) == ['a', 'args', 'b', 'kwargs']
    assert sig.parameters['a'].annotation == 'int'
    assert sig.parameters['args'].annotation == 'bytes'
    assert sig.parameters['b'].annotation == 'str'
    assert sig.parameters['kwargs'].annotation == 'float'
    assert sig.return_annotation == 'None'


def test_signature_from_str_complex_annotations() -> None:
    sig = inspect.signature_from_str('() -> Tuple[str, int, ...]')
    assert sig.return_annotation == 'Tuple[str, int, ...]'

    sig = inspect.signature_from_str('() -> Callable[[int, int], int]')
    assert sig.return_annotation == 'Callable[[int, int], int]'


def test_signature_from_str_kwonly_args() -> None:
    sig = inspect.signature_from_str('(a, *, b)')
    assert list(sig.parameters.keys()) == ['a', 'b']
    assert sig.parameters['a'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['a'].default == Parameter.empty
    assert sig.parameters['b'].kind == Parameter.KEYWORD_ONLY
    assert sig.parameters['b'].default == Parameter.empty


def test_signature_from_str_positionaly_only_args() -> None:
    sig = inspect.signature_from_str('(a, b=0, /, c=1)')
    assert list(sig.parameters.keys()) == ['a', 'b', 'c']
    assert sig.parameters['a'].kind == Parameter.POSITIONAL_ONLY
    assert sig.parameters['a'].default == Parameter.empty
    assert sig.parameters['b'].kind == Parameter.POSITIONAL_ONLY
    assert sig.parameters['b'].default == '0'
    assert sig.parameters['c'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['c'].default == '1'


def test_signature_from_str_invalid() -> None:
    with pytest.raises(SyntaxError):
        inspect.signature_from_str('')


def test_signature_from_ast():
    signature = 'def func(a, b, *args, c=0, d="blah", **kwargs): pass'
    tree = ast.parse(signature)
    sig = inspect.signature_from_ast(tree.body[0])
    assert list(sig.parameters.keys()) == ['a', 'b', 'args', 'c', 'd', 'kwargs']
    assert sig.parameters['a'].name == 'a'
    assert sig.parameters['a'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['a'].default == Parameter.empty
    assert sig.parameters['a'].annotation == Parameter.empty
    assert sig.parameters['b'].name == 'b'
    assert sig.parameters['b'].kind == Parameter.POSITIONAL_OR_KEYWORD
    assert sig.parameters['b'].default == Parameter.empty
    assert sig.parameters['b'].annotation == Parameter.empty
    assert sig.parameters['args'].name == 'args'
    assert sig.parameters['args'].kind == Parameter.VAR_POSITIONAL
    assert sig.parameters['args'].default == Parameter.empty
    assert sig.parameters['args'].annotation == Parameter.empty
    assert sig.parameters['c'].name == 'c'
    assert sig.parameters['c'].kind == Parameter.KEYWORD_ONLY
    assert sig.parameters['c'].default == '0'
    assert sig.parameters['c'].annotation == Parameter.empty
    assert sig.parameters['d'].name == 'd'
    assert sig.parameters['d'].kind == Parameter.KEYWORD_ONLY
    assert sig.parameters['d'].default == "'blah'"
    assert sig.parameters['d'].annotation == Parameter.empty
    assert sig.parameters['kwargs'].name == 'kwargs'
    assert sig.parameters['kwargs'].kind == Parameter.VAR_KEYWORD
    assert sig.parameters['kwargs'].default == Parameter.empty
    assert sig.parameters['kwargs'].annotation == Parameter.empty
    assert sig.return_annotation == Parameter.empty


def test_safe_getattr_with_default() -> None:
    class Foo:
        def __getattr__(self, item):
            raise Exception

    obj = Foo()

    result = inspect.safe_getattr(obj, 'bar', 'baz')

    assert result == 'baz'


def test_safe_getattr_with_exception() -> None:
    class Foo:
        def __getattr__(self, item):
            raise Exception

    obj = Foo()

    with pytest.raises(AttributeError, match='bar'):
        inspect.safe_getattr(obj, 'bar')


def test_safe_getattr_with_property_exception() -> None:
    class Foo:
        @property
        def bar(self):
            raise Exception

    obj = Foo()

    with pytest.raises(AttributeError, match='bar'):
        inspect.safe_getattr(obj, 'bar')


def test_safe_getattr_with___dict___override() -> None:
    class Foo:
        @property
        def __dict__(self):
            raise Exception

    obj = Foo()

    with pytest.raises(AttributeError, match='bar'):
        inspect.safe_getattr(obj, 'bar')


def test_dictionary_sorting() -> None:
    dictionary = {'c': 3, 'a': 1, 'd': 2, 'b': 4}
    description = inspect.object_description(dictionary)
    assert description == "{'a': 1, 'b': 4, 'c': 3, 'd': 2}"


def test_set_sorting() -> None:
    set_ = set('gfedcba')
    description = inspect.object_description(set_)
    assert description == "{'a', 'b', 'c', 'd', 'e', 'f', 'g'}"


def test_set_sorting_enum() -> None:
    class MyEnum(enum.Enum):
        a = 1
        b = 2
        c = 3

    set_ = set(MyEnum)
    description = inspect.object_description(set_)
    assert description == '{MyEnum.a, MyEnum.b, MyEnum.c}'


def test_set_sorting_fallback() -> None:
    set_ = {None, 1}
    description = inspect.object_description(set_)
    assert description == '{1, None}'


def test_deterministic_nested_collection_descriptions() -> None:
    # sortable
    assert inspect.object_description([{1, 2, 3, 10}]) == '[{1, 2, 3, 10}]'
    assert inspect.object_description(({1, 2, 3, 10},)) == '({1, 2, 3, 10},)'
    # non-sortable (elements of varying datatype)
    assert inspect.object_description([{None, 1}]) == '[{1, None}]'
    assert inspect.object_description(({None, 1},)) == '({1, None},)'
    assert inspect.object_description([{None, 1, 'A'}]) == "[{'A', 1, None}]"
    assert inspect.object_description(({None, 1, 'A'},)) == "({'A', 1, None},)"


def test_frozenset_sorting() -> None:
    frozenset_ = frozenset('gfedcba')
    description = inspect.object_description(frozenset_)
    assert description == "frozenset({'a', 'b', 'c', 'd', 'e', 'f', 'g'})"


def test_frozenset_sorting_fallback() -> None:
    frozenset_ = frozenset((None, 1))
    description = inspect.object_description(frozenset_)
    assert description == 'frozenset({1, None})'


def test_nested_tuple_sorting():
    tuple_ = ({'c', 'b', 'a'},)  # nb. trailing comma
    description = inspect.object_description(tuple_)
    assert description == "({'a', 'b', 'c'},)"

    tuple_ = ({'c', 'b', 'a'}, {'f', 'e', 'd'})
    description = inspect.object_description(tuple_)
    assert description == "({'a', 'b', 'c'}, {'d', 'e', 'f'})"


def test_recursive_collection_description():
    dict_a_, dict_b_ = {'a': 1}, {'b': 2}
    dict_a_['link'], dict_b_['link'] = dict_b_, dict_a_
    description_a, description_b = (
        inspect.object_description(dict_a_),
        inspect.object_description(dict_b_),
    )
    assert description_a == "{'a': 1, 'link': {'b': 2, 'link': dict(...)}}"
    assert description_b == "{'b': 2, 'link': {'a': 1, 'link': dict(...)}}"

    list_c_, list_d_ = [1, 2, 3, 4], [5, 6, 7, 8]
    list_c_.append(list_d_)
    list_d_.append(list_c_)
    description_c, description_d = (
        inspect.object_description(list_c_),
        inspect.object_description(list_d_),
    )

    assert description_c == '[1, 2, 3, 4, [5, 6, 7, 8, list(...)]]'
    assert description_d == '[5, 6, 7, 8, [1, 2, 3, 4, list(...)]]'


def test_dict_customtype() -> None:
    class CustomType:
        def __init__(self, value):
            self._value = value

        def __repr__(self):
            return '<CustomType(%r)>' % self._value

    dictionary = {CustomType(2): 2, CustomType(1): 1}
    description = inspect.object_description(dictionary)
    # Type is unsortable, just check that it does not crash
    assert '<CustomType(2)>: 2' in description


def test_object_description_enum() -> None:
    class MyEnum(enum.Enum):
        FOO = 1
        BAR = 2

    assert inspect.object_description(MyEnum.FOO) == 'MyEnum.FOO'


def test_object_description_enum_custom_repr() -> None:
    class MyEnum(enum.Enum):
        FOO = 1
        BAR = 2

        def __repr__(self):
            return self.name

    assert inspect.object_description(MyEnum.FOO) == 'FOO'


def test_getslots() -> None:
    class Foo:
        pass

    class Bar:
        __slots__ = ['attr']

    class Baz:
        __slots__ = {'attr': 'docstring'}

    class Qux:
        __slots__ = 'attr'  # NoQA: PLC0205

    assert inspect.getslots(Foo) is None
    assert inspect.getslots(Bar) == {'attr': None}
    assert inspect.getslots(Baz) == {'attr': 'docstring'}
    assert inspect.getslots(Qux) == {'attr': None}

    with pytest.raises(TypeError):
        inspect.getslots(Bar())


@pytest.mark.parametrize(
    ('expect', 'klass', 'name'),
    [
        # class methods
        (True, Base, 'classmeth'),
        (True, Inherited, 'classmeth'),
        (True, MyInt, 'classmeth'),
        (True, MyIntOverride, 'from_bytes'),
        # non class methods
        (False, Base, 'meth'),
        (False, Inherited, 'meth'),
        (False, MyInt, 'conjugate'),
        (False, MyIntOverride, 'conjugate'),
    ],
)
def test_isclassmethod(expect, klass, name):
    subject = getattr(klass, name)
    assert inspect.isclassmethod(subject) is expect
    assert inspect.isclassmethod(None, klass, name) is expect


@pytest.mark.parametrize(
    ('expect', 'klass', 'dict_key'),
    [
        # int.from_bytes is not a class method descriptor
        # but int.__dict__['from_bytes'] is one.
        (True, int, 'from_bytes'),
        (True, MyInt, 'from_bytes'),  # inherited
        # non class method descriptors
        (False, Base, 'classmeth'),
        (False, Inherited, 'classmeth'),
        (False, int, '__init__'),
        (False, int, 'conjugate'),
        (False, MyInt, 'classmeth'),
        (False, MyIntOverride, 'from_bytes'),  # overridden in pure Python
    ],
)
def test_is_classmethod_descriptor(expect, klass, dict_key):
    in_dict = dict_key in klass.__dict__
    subject = klass.__dict__.get(dict_key)
    assert inspect.is_classmethod_descriptor(subject) is (in_dict and expect)
    assert inspect.is_classmethod_descriptor(None, klass, dict_key) is expect

    method = getattr(klass, dict_key)
    assert not inspect.is_classmethod_descriptor(method)


@pytest.mark.parametrize(
    ('expect', 'klass', 'dict_key'),
    [
        # class method descriptors
        (True, int, 'from_bytes'),
        (True, bytes, 'fromhex'),
        (True, MyInt, 'from_bytes'),  # in C only
        # non class method descriptors
        (False, Base, 'classmeth'),
        (False, Inherited, 'classmeth'),
        (False, int, '__init__'),
        (False, int, 'conjugate'),
        (False, MyInt, 'classmeth'),
        (False, MyIntOverride, 'from_bytes'),  # overridden in pure Python
    ],
)
def test_is_builtin_classmethod_like(expect, klass, dict_key):
    method = getattr(klass, dict_key)
    assert inspect.is_builtin_classmethod_like(method) is expect
    assert inspect.is_builtin_classmethod_like(None, klass, dict_key) is expect


@pytest.mark.parametrize(
    ('expect', 'klass', 'name'),
    [
        # regular class methods
        (True, Base, 'classmeth'),
        (True, Inherited, 'classmeth'),
        (True, MyInt, 'classmeth'),
        # inherited C class method
        (True, MyIntOverride, 'from_bytes'),
        # class method descriptors
        (True, int, 'from_bytes'),
        (True, bytes, 'fromhex'),
        (True, MyInt, 'from_bytes'),  # in C only
        # not classmethod-like
        (False, int, '__init__'),
        (False, int, 'conjugate'),
        (False, MyIntOverride, 'conjugate'),  # overridden in pure Python
    ],
)
def test_is_classmethod_like(expect, klass, name):
    subject = getattr(klass, name)
    assert inspect.is_classmethod_like(subject) is expect


def test_isstaticmethod() -> None:
    assert inspect.isstaticmethod(Base.staticmeth, Base, 'staticmeth')
    assert not inspect.isstaticmethod(Base.meth, Base, 'meth')
    assert inspect.isstaticmethod(Inherited.staticmeth, Inherited, 'staticmeth')
    assert not inspect.isstaticmethod(Inherited.meth, Inherited, 'meth')


def test_iscoroutinefunction() -> None:
    # function
    assert not inspect.iscoroutinefunction(func)
    # coroutine
    assert inspect.iscoroutinefunction(coroutinefunc)
    # partial-ed coroutine
    assert inspect.iscoroutinefunction(partial_coroutinefunc)
    # method
    assert not inspect.iscoroutinefunction(Base.meth)
    # coroutine-method
    assert inspect.iscoroutinefunction(Base.coroutinemeth)
    # coroutine classmethod
    assert inspect.iscoroutinefunction(Base.__dict__['coroutineclassmeth'])

    # partial-ed coroutine-method
    partial_coroutinemeth = Base.__dict__['partial_coroutinemeth']
    assert inspect.iscoroutinefunction(partial_coroutinemeth)


def test_iscoroutinefunction_wrapped() -> None:
    # function wrapping a callable obj
    assert inspect.isfunction(_decorator(coroutinefunc))


def test_isfunction() -> None:
    # fmt: off
    assert inspect.isfunction(func)                      # function
    assert inspect.isfunction(partial_func)              # partial-ed function
    assert inspect.isfunction(Base.meth)                 # method of class
    assert inspect.isfunction(Base.partialmeth)          # partial-ed method of class
    assert not inspect.isfunction(Base().meth)           # method of instance
    assert not inspect.isfunction(builtin_func)          # builtin function
    assert not inspect.isfunction(partial_builtin_func)  # partial-ed builtin function
    # fmt: on


def test_isfunction_wrapped() -> None:
    # function wrapping a callable obj
    assert inspect.isfunction(_decorator(_Callable()))


def test_isbuiltin() -> None:
    # fmt: off
    assert inspect.isbuiltin(builtin_func)          # builtin function
    assert inspect.isbuiltin(partial_builtin_func)  # partial-ed builtin function
    assert not inspect.isbuiltin(func)              # function
    assert not inspect.isbuiltin(partial_func)      # partial-ed function
    assert not inspect.isbuiltin(Base.meth)         # method of class
    assert not inspect.isbuiltin(Base().meth)       # method of instance
    # fmt: on


def test_isdescriptor() -> None:
    # fmt: off
    assert inspect.isdescriptor(Base.prop)        # property of class
    assert not inspect.isdescriptor(Base().prop)  # property of instance
    assert inspect.isdescriptor(Base.meth)        # method of class
    assert inspect.isdescriptor(Base().meth)      # method of instance
    assert inspect.isdescriptor(func)             # function
    # fmt: on


def test_isattributedescriptor() -> None:
    # fmt: off
    assert inspect.isattributedescriptor(Base.prop)                      # property
    assert not inspect.isattributedescriptor(Base.meth)                  # method
    assert not inspect.isattributedescriptor(Base.staticmeth)            # staticmethod
    assert not inspect.isattributedescriptor(Base.classmeth)             # classmetho
    assert not inspect.isattributedescriptor(Descriptor)                 # custom descriptor class
    assert not inspect.isattributedescriptor(str.join)                   # MethodDescriptorType
    assert not inspect.isattributedescriptor(object.__init__)            # WrapperDescriptorType
    assert not inspect.isattributedescriptor(dict.__dict__['fromkeys'])  # ClassMethodDescriptorType
    assert inspect.isattributedescriptor(types.FrameType.f_locals)       # GetSetDescriptorType
    assert inspect.isattributedescriptor(datetime.timedelta.days)        # MemberDescriptorType
    # fmt: on

    try:
        # _testcapi module cannot be importable in some distro
        # See: https://github.com/sphinx-doc/sphinx/issues/9868
        import _testcapi  # type: ignore[import-not-found]

        # instancemethod (C-API)
        testinstancemethod = _testcapi.instancemethod(str.__repr__)
        assert not inspect.isattributedescriptor(testinstancemethod)
    except ImportError:
        pass


def test_isproperty() -> None:
    # fmt: off
    assert inspect.isproperty(Base.prop)        # property of class
    assert not inspect.isproperty(Base().prop)  # property of instance
    assert not inspect.isproperty(Base.meth)    # method of class
    assert not inspect.isproperty(Base().meth)  # method of instance
    assert not inspect.isproperty(func)         # function
    # fmt: on


def test_isgenericalias() -> None:
    #: A list of int
    T = List[int]  # NoQA: UP006
    S = list[Union[str, None]]  # NoQA: UP007

    C = Callable[[int], None]  # a generic alias not having a doccomment

    assert inspect.isgenericalias(C)
    assert inspect.isgenericalias(Callable)
    assert inspect.isgenericalias(T)
    assert inspect.isgenericalias(List)  # NoQA: UP006
    assert inspect.isgenericalias(S)
    assert not inspect.isgenericalias(list)
    assert not inspect.isgenericalias([])
    assert not inspect.isgenericalias(object())
    assert not inspect.isgenericalias(Base)


def test_unpartial() -> None:
    def func1(a, b, c):
        pass

    func2 = functools.partial(func1, 1)
    func2.__doc__ = 'func2'
    func3 = functools.partial(func2, 2)  # nested partial object

    assert inspect.unpartial(func2) is func1
    assert inspect.unpartial(func3) is func1


def test_getdoc_inherited_classmethod():
    class Foo:
        @classmethod
        def meth(cls):
            """Docstring
            indented text
            """

    class Bar(Foo):
        @classmethod
        def meth(cls):
            # inherited classmethod
            pass

    assert inspect.getdoc(Bar.meth, getattr, False, Bar, 'meth') is None
    assert inspect.getdoc(Bar.meth, getattr, True, Bar, 'meth') == Foo.meth.__doc__


def test_getdoc_inherited_decorated_method():
    class Foo:
        def meth(self):
            """Docstring
            indented text
            """

    class Bar(Foo):
        @functools.lru_cache  # NoQA: B019
        def meth(self):
            # inherited and decorated method
            pass

    assert inspect.getdoc(Bar.meth, getattr, False, Bar, 'meth') is None
    assert inspect.getdoc(Bar.meth, getattr, True, Bar, 'meth') == Foo.meth.__doc__


def test_is_builtin_class_method() -> None:
    class MyInt(int):
        def my_method(self):
            pass

    assert inspect.is_builtin_class_method(MyInt, 'to_bytes')
    assert inspect.is_builtin_class_method(MyInt, '__init__')
    assert not inspect.is_builtin_class_method(MyInt, 'my_method')
    assert not inspect.is_builtin_class_method(MyInt, 'does_not_exist')
    assert not inspect.is_builtin_class_method(4, 'still does not crash')

    class ObjectWithMroAttr:
        def __init__(self, mro_attr):
            self.__mro__ = mro_attr

    assert not inspect.is_builtin_class_method(
        ObjectWithMroAttr([1, 2, 3]), 'still does not crash'
    )



"""Test sphinx.ext.duration extension."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'dummy',
    testroot='basic',
    confoverrides={'extensions': ['sphinx.ext.duration']},
)
def test_githubpages(app: SphinxTestApp) -> None:
    app.build()

    assert 'slowest reading durations' in app.status.getvalue()
    assert re.search('\\d+\\.\\d{3} index\n', app.status.getvalue())



"""Test the build process with Text builder with the test root."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils.utils import column_width

from sphinx.testing.util import SphinxTestApp
from sphinx.writers.text import MAXWIDTH, Cell, Table

if TYPE_CHECKING:
    from typing import Any

    from sphinx.testing.util import SphinxTestApp


def with_text_app(*args: Any, **kw: Any) -> pytest.MarkDecorator:
    return pytest.mark.sphinx(*args, buildername='text', testroot='build-text', **kw)


@with_text_app()
def test_maxwitdh_with_prefix(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'maxwidth.txt').read_text(encoding='utf8')

    lines = result.splitlines()
    line_widths = [column_width(line) for line in lines]
    assert max(line_widths) < MAXWIDTH
    assert lines[0].startswith('See also:')
    assert lines[1].startswith('')
    assert lines[2].startswith('  ham')
    assert lines[3].startswith('  ham')
    assert lines[4] == ''
    assert lines[5].startswith('* ham')
    assert lines[6].startswith('  ham')
    assert lines[7] == ''
    assert lines[8].startswith('* ham')
    assert lines[9].startswith('  ham')
    assert lines[10] == ''
    assert lines[11].startswith('spam egg')


@with_text_app()
def test_lineblock(app: SphinxTestApp) -> None:
    # regression test for https://github.com/sphinx-doc/sphinx/issues/1109:
    # need empty line after line block
    app.build()
    result = (app.outdir / 'lineblock.txt').read_text(encoding='utf8')
    expect = '* one\n\n     line-block 1\n     line-block 2\n\nfollowed paragraph.\n'
    assert result == expect


@with_text_app()
def test_nonascii_title_line(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    result = (app.outdir / 'nonascii_title.txt').read_text(encoding='utf8')
    expect_underline = '*********'
    result_underline = result.splitlines()[1].strip()
    assert result_underline == expect_underline


@with_text_app()
def test_nonascii_table(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'nonascii_table.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    line_widths = [column_width(line) for line in lines]
    assert len(set(line_widths)) == 1  # same widths


@with_text_app()
def test_nonascii_maxwidth(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'nonascii_maxwidth.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    line_widths = [column_width(line) for line in lines]
    assert max(line_widths) < MAXWIDTH


def test_table_builder() -> None:
    table = Table([6, 6])
    table.add_cell(Cell('foo'))
    table.add_cell(Cell('bar'))
    table_str = str(table).split('\n')
    assert table_str[0] == '+--------+--------+'
    assert table_str[1] == '| foo    | bar    |'
    assert table_str[2] == '+--------+--------+'
    assert repr(table).count('<Cell ') == 2


def test_table_separator() -> None:
    table = Table([6, 6])
    table.add_cell(Cell('foo'))
    table.add_cell(Cell('bar'))
    table.set_separator()
    table.add_row()
    table.add_cell(Cell('FOO'))
    table.add_cell(Cell('BAR'))
    table_str = str(table).split('\n')
    assert table_str[0] == '+--------+--------+'
    assert table_str[1] == '| foo    | bar    |'
    assert table_str[2] == '|========|========|'
    assert table_str[3] == '| FOO    | BAR    |'
    assert table_str[4] == '+--------+--------+'
    assert repr(table).count('<Cell ') == 4


def test_table_cell() -> None:
    cell = Cell('Foo bar baz')
    cell.wrap(3)
    assert 'Cell' in repr(cell)
    assert cell.wrapped == ['Foo', 'bar', 'baz']


@with_text_app()
def test_table_with_empty_cell(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'table.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    assert lines[0] == '+-------+-------+'
    assert lines[1] == '| XXX   | XXX   |'
    assert lines[2] == '+-------+-------+'
    assert lines[3] == '|       | XXX   |'
    assert lines[4] == '+-------+-------+'
    assert lines[5] == '| XXX   |       |'
    assert lines[6] == '+-------+-------+'


@with_text_app()
def test_table_with_rowspan(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'table_rowspan.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    assert lines[0] == '+-------+-------+'
    assert lines[1] == '| XXXXXXXXX     |'
    assert lines[2] == '+-------+-------+'
    assert lines[3] == '|       | XXX   |'
    assert lines[4] == '+-------+-------+'
    assert lines[5] == '| XXX   |       |'
    assert lines[6] == '+-------+-------+'


@with_text_app()
def test_table_with_colspan(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'table_colspan.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    assert lines[0] == '+-------+-------+'
    assert lines[1] == '| XXX   | XXX   |'
    assert lines[2] == '+-------+-------+'
    assert lines[3] == '|       | XXX   |'
    assert lines[4] == '+-------+       |'
    assert lines[5] == '| XXX   |       |'
    assert lines[6] == '+-------+-------+'


@with_text_app()
def test_table_with_colspan_left(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'table_colspan_left.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    assert lines[0] == '+-------+-------+'
    assert lines[1] == '| XXX   | XXX   |'
    assert lines[2] == '+-------+-------+'
    assert lines[3] == '| XXX   | XXX   |'
    assert lines[4] == '|       +-------+'
    assert lines[5] == '|       |       |'
    assert lines[6] == '+-------+-------+'


@with_text_app()
def test_table_with_colspan_and_rowspan(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'table_colspan_and_rowspan.txt').read_text(encoding='utf8')
    lines = [line.strip() for line in result.splitlines() if line.strip()]
    assert result
    assert lines[0] == '+-------+-------+-------+'
    assert lines[1] == '| AAA           | BBB   |'
    assert lines[2] == '+-------+-------+       |'
    assert lines[3] == '| DDD   | XXX   |       |'
    assert lines[4] == '|       +-------+-------+'
    assert lines[5] == '|       | CCC           |'
    assert lines[6] == '+-------+-------+-------+'


@with_text_app()
def test_list_items_in_admonition(app: SphinxTestApp) -> None:
    app.build()
    result = (app.outdir / 'listitems.txt').read_text(encoding='utf8')
    lines = [line.rstrip() for line in result.splitlines()]
    assert lines[0] == 'See also:'
    assert lines[1] == ''
    assert lines[2] == '  * item 1'
    assert lines[3] == ''
    assert lines[4] == '  * item 2'


@with_text_app()
def test_secnums(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    index = (app.outdir / 'index.txt').read_text(encoding='utf8')
    lines = index.splitlines()
    assert lines[0] == '* 1. Section A'
    assert lines[1] == ''
    assert lines[2] == '* 2. Section B'
    assert lines[3] == ''
    assert lines[4] == '  * 2.1. Sub Ba'
    assert lines[5] == ''
    assert lines[6] == '  * 2.2. Sub Bb'
    doc2 = (app.outdir / 'doc2.txt').read_text(encoding='utf8')
    expect = (
        '2. Section B\n'
        '************\n'
        '\n'
        '\n'
        '2.1. Sub Ba\n'
        '===========\n'
        '\n'
        '\n'
        '2.2. Sub Bb\n'
        '===========\n'
    )
    assert doc2 == expect

    app.config.text_secnumber_suffix = ' '
    app.build(force_all=True)
    index = (app.outdir / 'index.txt').read_text(encoding='utf8')
    lines = index.splitlines()
    assert lines[0] == '* 1 Section A'
    assert lines[1] == ''
    assert lines[2] == '* 2 Section B'
    assert lines[3] == ''
    assert lines[4] == '  * 2.1 Sub Ba'
    assert lines[5] == ''
    assert lines[6] == '  * 2.2 Sub Bb'
    doc2 = (app.outdir / 'doc2.txt').read_text(encoding='utf8')
    expect = (
        '2 Section B\n'
        '***********\n'
        '\n'
        '\n'
        '2.1 Sub Ba\n'
        '==========\n'
        '\n'
        '\n'
        '2.2 Sub Bb\n'
        '==========\n'
    )
    assert doc2 == expect

    app.config.text_add_secnumbers = False
    app.build(force_all=True)
    index = (app.outdir / 'index.txt').read_text(encoding='utf8')
    lines = index.splitlines()
    assert lines[0] == '* Section A'
    assert lines[1] == ''
    assert lines[2] == '* Section B'
    assert lines[3] == ''
    assert lines[4] == '  * Sub Ba'
    assert lines[5] == ''
    assert lines[6] == '  * Sub Bb'
    doc2 = (app.outdir / 'doc2.txt').read_text(encoding='utf8')
    expect = 'Section B\n*********\n\n\nSub Ba\n======\n\n\nSub Bb\n======\n'
    assert doc2 == expect



from __future__ import annotations

from docutils import nodes

from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective, SphinxRole
from sphinx.util.typing import ExtensionMetadata


class HelloRole(SphinxRole):
    """A role to say hello!"""

    def run(self) -> tuple[list[nodes.Node], list[nodes.system_message]]:
        node = nodes.inline(text=f'Hello {self.text}!')
        return [node], []


class HelloDirective(SphinxDirective):
    """A directive to say hello!"""

    required_arguments = 1

    def run(self) -> list[nodes.Node]:
        paragraph_node = nodes.paragraph(text=f'hello {self.arguments[0]}!')
        return [paragraph_node]


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_role('hello', HelloRole())
    app.add_directive('hello', HelloDirective)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from __future__ import annotations

import pytest

from sphinx.util._lines import parse_line_num_spec


def test_parse_line_num_spec() -> None:
    assert parse_line_num_spec('1,2,3', 10) == [0, 1, 2]
    assert parse_line_num_spec('4, 5, 6', 10) == [3, 4, 5]
    assert parse_line_num_spec('-4', 10) == [0, 1, 2, 3]
    assert parse_line_num_spec('7-9', 10) == [6, 7, 8]
    assert parse_line_num_spec('7-', 10) == [6, 7, 8, 9]
    assert parse_line_num_spec('1,7-', 10) == [0, 6, 7, 8, 9]
    assert parse_line_num_spec('7-7', 10) == [6]
    assert parse_line_num_spec('11-', 10) == [10]
    with pytest.raises(ValueError, match="invalid line number spec: '1-2-3'"):
        parse_line_num_spec('1-2-3', 10)
    with pytest.raises(ValueError, match="invalid line number spec: 'abc-def'"):
        parse_line_num_spec('abc-def', 10)
    with pytest.raises(ValueError, match="invalid line number spec: '-'"):
        parse_line_num_spec('-', 10)
    with pytest.raises(ValueError, match="invalid line number spec: '3-1'"):
        parse_line_num_spec('3-1', 10)



"""mod1"""


def decorator(f):
    return f


@decorator
def func1(a, b):
    """this is func1"""
    return a, b


@decorator
class Class1:
    """this is Class1"""


class Class3:
    """this is Class3"""

    class_attr = 42
    """this is the class attribute class_attr"""



"""Test sphinx.util._importer."""

from __future__ import annotations

import pytest

from sphinx.errors import ExtensionError
from sphinx.util._importer import import_object


def test_import_object() -> None:
    module = import_object('sphinx')
    assert module.__name__ == 'sphinx'

    module = import_object('sphinx.application')
    assert module.__name__ == 'sphinx.application'

    obj = import_object('sphinx.application.Sphinx')
    assert obj.__name__ == 'Sphinx'

    with pytest.raises(ExtensionError) as exc:
        import_object('sphinx.unknown_module')
    assert exc.value.args[0] == 'Could not import sphinx.unknown_module'

    with pytest.raises(ExtensionError) as exc:
        import_object('sphinx.unknown_module', 'my extension')
    expected = 'Could not import sphinx.unknown_module (needed for my extension)'
    assert exc.value.args[0] == expected



"""Test pycode."""

from __future__ import annotations

import sys
from importlib.machinery import SourceFileLoader
from pathlib import Path

import pytest

import sphinx
from sphinx.errors import PycodeError
from sphinx.pycode import ModuleAnalyzer

SPHINX_MODULE_PATH = Path(sphinx.__file__).resolve().with_suffix('.py')


def test_ModuleAnalyzer_get_module_source() -> None:
    assert isinstance(sphinx.__spec__.loader, SourceFileLoader)  # type checking
    assert ModuleAnalyzer.get_module_source('sphinx') == (
        Path(sphinx.__file__),
        sphinx.__spec__.loader.get_source('sphinx'),
    )

    # failed to obtain source information from builtin modules
    with pytest.raises(PycodeError):
        ModuleAnalyzer.get_module_source('builtins')
    with pytest.raises(PycodeError):
        ModuleAnalyzer.get_module_source('itertools')


def test_ModuleAnalyzer_for_string() -> None:
    analyzer = ModuleAnalyzer.for_string('print("Hello world")', 'module_name')
    assert analyzer.modname == 'module_name'
    assert analyzer.srcname == '<string>'


def test_ModuleAnalyzer_for_file() -> None:
    analyzer = ModuleAnalyzer.for_file(SPHINX_MODULE_PATH, 'sphinx')
    assert analyzer.modname == 'sphinx'
    assert analyzer.srcname == str(SPHINX_MODULE_PATH)


def test_ModuleAnalyzer_for_module(rootdir):
    analyzer = ModuleAnalyzer.for_module('sphinx')
    assert analyzer.modname == 'sphinx'
    assert analyzer.srcname == str(SPHINX_MODULE_PATH)

    saved_path = sys.path.copy()
    sys.path.insert(0, str(rootdir / 'test-pycode'))
    try:
        analyzer = ModuleAnalyzer.for_module('cp_1251_coded')
        docs = analyzer.find_attr_docs()
        assert docs == {('', 'X'): ['It MUST look like X="\u0425"', '']}
    finally:
        sys.path[:] = saved_path


def test_ModuleAnalyzer_find_tags() -> None:
    code = (
        'class Foo(object):\n'  # line: 1
        '    """class Foo!"""\n'
        '    def __init__(self):\n'
        '       pass\n'
        '\n'
        '    def bar(self, arg1, arg2=True, *args, **kwargs):\n'
        '       """method Foo.bar"""\n'
        '       pass\n'
        '\n'
        '    class Baz(object):\n'
        '       def __init__(self):\n'  # line: 11
        '           pass\n'
        '\n'
        'def qux():\n'
        '   """function baz"""\n'
        '   pass\n'
        '\n'
        '@decorator1\n'
        '@decorator2\n'
        'def quux():\n'
        '   pass\n'  # line: 21
        '\n'
        'class Corge(object):\n'
        '    @decorator1\n'
        '    @decorator2\n'
        '    def grault(self):\n'
        '        pass\n'
    )
    analyzer = ModuleAnalyzer.for_string(code, 'module')
    tags = analyzer.find_tags()
    assert set(tags.keys()) == {
        'Foo',
        'Foo.__init__',
        'Foo.bar',
        'Foo.Baz',
        'Foo.Baz.__init__',
        'qux',
        'quux',
        'Corge',
        'Corge.grault',
    }
    assert tags['Foo'] == ('class', 1, 12)  # type, start, end
    assert tags['Foo.__init__'] == ('def', 3, 4)
    assert tags['Foo.bar'] == ('def', 6, 8)
    assert tags['Foo.Baz'] == ('class', 10, 12)
    assert tags['Foo.Baz.__init__'] == ('def', 11, 12)
    assert tags['qux'] == ('def', 14, 16)
    assert tags['quux'] == ('def', 18, 21)
    assert tags['Corge'] == ('class', 23, 27)
    assert tags['Corge.grault'] == ('def', 24, 27)


def test_ModuleAnalyzer_find_attr_docs() -> None:
    code = (
        'class Foo(object):\n'
        '    """class Foo!"""\n'
        '    #: comment before attr1\n'
        '    attr1 = None\n'
        '    attr2 = None  # attribute comment for attr2 (without colon)\n'
        '    attr3 = None  #: attribute comment for attr3\n'
        '    attr4 = None  #: long attribute comment\n'
        '                  #: for attr4\n'
        '    #: comment before attr5\n'
        '    attr5 = None  #: attribute comment for attr5\n'
        '    attr6, attr7 = 1, 2  #: this comment is ignored\n'
        '\n'
        '    def __init__(self):\n'
        '       self.attr8 = None  #: first attribute comment (ignored)\n'
        '       self.attr8 = None  #: attribute comment for attr8\n'
        '       #: comment before attr9\n'
        '       self.attr9 = None  #: comment after attr9\n'
        '       "string after attr9"\n'
        '\n'
        '    def bar(self, arg1, arg2=True, *args, **kwargs):\n'
        '       """method Foo.bar"""\n'
        '       pass\n'
        '\n'
        'def baz():\n'
        '   """function baz"""\n'
        '   pass\n'
        '\n'
        'class Qux: attr1 = 1; attr2 = 2'
    )
    analyzer = ModuleAnalyzer.for_string(code, 'module')
    docs = analyzer.find_attr_docs()
    assert set(docs) == {
        ('Foo', 'attr1'),
        ('Foo', 'attr3'),
        ('Foo', 'attr4'),
        ('Foo', 'attr5'),
        ('Foo', 'attr6'),
        ('Foo', 'attr7'),
        ('Foo', 'attr8'),
        ('Foo', 'attr9'),
    }
    assert docs['Foo', 'attr1'] == ['comment before attr1', '']
    assert docs['Foo', 'attr3'] == ['attribute comment for attr3', '']
    assert docs['Foo', 'attr4'] == ['long attribute comment', '']
    assert docs['Foo', 'attr4'] == ['long attribute comment', '']
    assert docs['Foo', 'attr5'] == ['attribute comment for attr5', '']
    assert docs['Foo', 'attr6'] == ['this comment is ignored', '']
    assert docs['Foo', 'attr7'] == ['this comment is ignored', '']
    assert docs['Foo', 'attr8'] == ['attribute comment for attr8', '']
    assert docs['Foo', 'attr9'] == ['string after attr9', '']
    assert analyzer.tagorder == {
        'Foo': 0,
        'Foo.__init__': 8,
        'Foo.attr1': 1,
        'Foo.attr2': 2,
        'Foo.attr3': 3,
        'Foo.attr4': 4,
        'Foo.attr5': 5,
        'Foo.attr6': 6,
        'Foo.attr7': 7,
        'Foo.attr8': 10,
        'Foo.attr9': 12,
        'Foo.bar': 13,
        'baz': 14,
        'Qux': 15,
        'Qux.attr1': 16,
        'Qux.attr2': 17,
    }


def test_ModuleAnalyzer_find_attr_docs_for_posonlyargs_method() -> None:
    code = (
        'class Foo(object):\n'
        '    def __init__(self, /):\n'
        '       self.attr = None  #: attribute comment\n'
    )
    analyzer = ModuleAnalyzer.for_string(code, 'module')
    docs = analyzer.find_attr_docs()
    assert set(docs) == {('Foo', 'attr')}
    assert docs['Foo', 'attr'] == ['attribute comment', '']
    assert analyzer.tagorder == {'Foo': 0, 'Foo.__init__': 1, 'Foo.attr': 2}



from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import docutils
import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx('html', testroot='images')
def test_html_remote_images(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<img alt="http://localhost:7777/sphinx.png" '
        'src="http://localhost:7777/sphinx.png" />'
    ) in result
    assert not (app.outdir / 'sphinx.png').exists()


@pytest.mark.sphinx('html', testroot='image-escape')
def test_html_encoded_image(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<img alt="_images/img_%231.png" src="_images/img_%231.png" />' in result
    assert (app.outdir / '_images/img_#1.png').exists()


@pytest.mark.sphinx('html', testroot='remote-logo')
def test_html_remote_logo(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<img class="logo" src="https://www.python.org/static/img/python-logo.png" alt="Logo of Project name not set"/>'
    ) in result
    assert (
        '<link rel="icon" href="https://www.python.org/static/favicon.ico"/>'
    ) in result
    assert not (app.outdir / 'python-logo.png').exists()


@pytest.mark.sphinx('html', testroot='local-logo')
def test_html_local_logo(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<img class="logo" src="_static/img.png" alt="Logo of Project name not set"/>'
    ) in result
    assert (app.outdir / '_static/img.png').exists()


@pytest.mark.sphinx('html', testroot='html_scaled_image_link')
def test_html_scaled_image_link(app: SphinxTestApp) -> None:
    app.build()
    context = (app.outdir / 'index.html').read_text(encoding='utf8')

    # no scaled parameters
    assert re.search('\n<img alt="_images/img.png" src="_images/img.png" />', context)

    # scaled_image_link
    if docutils.__version_info__[:2] >= (0, 22):
        assert re.search(
            '\n<a class="reference internal image-reference" href="_images/img.png">'
            '<img alt="_images/img.png" height="[^"]+" src="_images/img.png" width="[^"]+" />'
            '\n</a>',
            context,
        )
    else:
        # Docutils 0.21 adds a newline before the closing </a> tag
        closing_space = '\n' if docutils.__version_info__[:2] >= (0, 21) else ''
        assert re.search(
            '\n<a class="reference internal image-reference" href="_images/img.png">'
            '<img alt="_images/img.png" src="_images/img.png" style="[^"]+" />'
            f'{closing_space}</a>',
            context,
        )

    # no-scaled-link class disables the feature
    if docutils.__version_info__[:2] >= (0, 22):
        assert re.search(
            '\n<img alt="_images/img.png" class="no-scaled-link"'
            ' height="[^"]+" src="_images/img.png" width="[^"]+" />',
            context,
        )
    else:
        assert re.search(
            '\n<img alt="_images/img.png" class="no-scaled-link"'
            ' src="_images/img.png" style="[^"]+" />',
            context,
        )


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx('html', testroot='images')
def test_copy_images(app: SphinxTestApp) -> None:
    app.build()

    images_dir = Path(app.outdir) / '_images'
    images = {image.name for image in images_dir.rglob('*')}
    assert images == {
        # 'ba30773957c3fe046897111afd65a80b81cad089.png',  # html: image from data:image/png URI in source
        'img.png',
        'rimg.png',
        'rimg1.png',
        'svgimg.svg',
        'testimäge.png',
    }



"""Tests util functions."""

from __future__ import annotations

import pytest

import sphinx.util
from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.deprecation import RemovedInSphinx10Warning, RemovedInSphinx90Warning
from sphinx.errors import ExtensionError
from sphinx.util._files import DownloadFiles, FilenameUniqDict
from sphinx.util._importer import import_object
from sphinx.util._lines import parse_line_num_spec
from sphinx.util._uri import encode_uri, is_url
from sphinx.util.index_entries import _split_into, split_index_msg
from sphinx.util.matching import patfilter
from sphinx.util.nodes import (
    caption_ref_re,
    explicit_title_re,
    nested_parse_with_titles,
    split_explicit_title,
)
from sphinx.util.osutil import (
    SEP,
    copyfile,
    ensuredir,
    make_filename,
    os_path,
    relative_uri,
)


def test_ensuredir(tmp_path):
    # Does not raise an exception for an existing directory.
    ensuredir(tmp_path)

    path = tmp_path / 'a' / 'b' / 'c'
    ensuredir(path)
    assert path.is_dir()


def test_exported_attributes() -> None:
    # RemovedInSphinx90Warning
    with pytest.warns(
        RemovedInSphinx90Warning,
        match=r"deprecated, use 'sphinx.util.index_entries.split_index_msg' instead.",
    ):
        assert sphinx.util.split_index_msg is split_index_msg
    with pytest.warns(RemovedInSphinx90Warning, match=r'deprecated.'):
        assert sphinx.util.split_into is _split_into
    with pytest.warns(
        RemovedInSphinx90Warning,
        match=r"deprecated, use 'sphinx.errors.ExtensionError' instead.",
    ):
        assert sphinx.util.ExtensionError is ExtensionError
    with pytest.warns(
        RemovedInSphinx90Warning,
        match=r"deprecated, use 'hashlib.md5' instead.",
    ):
        _ = sphinx.util.md5
    with pytest.warns(
        RemovedInSphinx90Warning,
        match=r"deprecated, use 'hashlib.sha1' instead.",
    ):
        _ = sphinx.util.sha1

    # RemovedInSphinx10Warning
    with pytest.warns(RemovedInSphinx10Warning, match=r'deprecated.'):
        assert sphinx.util.FilenameUniqDict is FilenameUniqDict
    with pytest.warns(RemovedInSphinx10Warning, match=r'deprecated.'):
        assert sphinx.util.DownloadFiles is DownloadFiles

    # Re-exported for backwards compatibility,
    # but not currently deprecated
    assert sphinx.util.encode_uri is encode_uri
    assert sphinx.util.import_object is import_object
    assert sphinx.util.isurl is is_url
    assert sphinx.util.parselinenos is parse_line_num_spec
    assert sphinx.util.patfilter is patfilter
    assert sphinx.util.strip_escape_sequences is strip_escape_sequences
    assert sphinx.util.caption_ref_re is caption_ref_re
    assert sphinx.util.explicit_title_re is explicit_title_re
    assert sphinx.util.nested_parse_with_titles is nested_parse_with_titles
    assert sphinx.util.split_explicit_title is split_explicit_title
    assert sphinx.util.SEP is SEP
    assert sphinx.util.copyfile is copyfile
    assert sphinx.util.ensuredir is ensuredir
    assert sphinx.util.make_filename is make_filename
    assert sphinx.util.os_path is os_path
    assert sphinx.util.relative_uri is relative_uri



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from typing import Any


@overload
def sum(x: int, y: int = 0) -> int: ...


@overload
def sum(x: float, y: float = 0.0) -> float: ...


@overload
def sum(x: str, y: str = ...) -> str: ...


def sum(x, y=None):
    """docstring"""
    return x + y


class Math:
    """docstring"""

    @overload
    def sum(self, x: int, y: int = 0) -> int: ...

    @overload
    def sum(self, x: float, y: float = 0.0) -> float: ...

    @overload
    def sum(self, x: str, y: str = ...) -> str: ...

    def sum(self, x, y=None):
        """docstring"""
        return x + y


class Foo:
    """docstring"""

    @overload
    def __new__(cls, x: int, y: int) -> Foo: ...

    @overload
    def __new__(cls, x: str, y: str) -> Foo: ...

    def __new__(cls, x, y):
        pass


class Bar:
    """docstring"""

    @overload
    def __init__(self, x: int, y: int) -> None: ...

    @overload
    def __init__(self, x: str, y: str) -> None: ...

    def __init__(self, x, y):
        pass


class Meta(type):
    @overload
    def __call__(cls, x: int, y: int) -> Any: ...

    @overload
    def __call__(cls, x: str, y: str) -> Any: ...

    def __call__(cls, x, y):
        pass


class Baz(metaclass=Meta):
    """docstring"""



"""Tests parsers module."""

from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from sphinx.parsers import RSTParser
from sphinx.util.docutils import new_document


@pytest.mark.sphinx('html', testroot='basic')
@patch('docutils.parsers.rst.states.RSTStateMachine')
def test_RSTParser_prolog_epilog(RSTStateMachine, app):
    document = new_document('dummy.rst')
    document.settings = Mock(tab_width=8, language_code='')
    parser = RSTParser()
    parser.set_application(app)

    # normal case
    text = 'hello Sphinx world\nSphinx is a document generator'
    parser.parse(text, document)
    (content, _), _ = RSTStateMachine().run.call_args

    assert list(content.xitems()) == [
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
    ]

    # with rst_prolog
    app.config.rst_prolog = 'this is rst_prolog\nhello reST!'
    parser.parse(text, document)
    (content, _), _ = RSTStateMachine().run.call_args
    assert list(content.xitems()) == [
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
    ]

    # with rst_epilog
    app.config.rst_prolog = None
    app.config.rst_epilog = 'this is rst_epilog\ngood-bye reST!'
    parser.parse(text, document)
    (content, _), _ = RSTStateMachine().run.call_args
    assert list(content.xitems()) == [
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
        ('dummy.rst', 2, ''),
        ('<rst_epilog>', 0, 'this is rst_epilog'),
        ('<rst_epilog>', 1, 'good-bye reST!'),
    ]

    # expandtabs / convert whitespaces
    app.config.rst_prolog = None
    app.config.rst_epilog = None
    text = '\thello Sphinx world\n\v\fSphinx is a document generator'
    parser.parse(text, document)
    (content, _), _ = RSTStateMachine().run.call_args
    assert list(content.xitems()) == [
        ('dummy.rst', 0, '        hello Sphinx world'),
        ('dummy.rst', 1, '  Sphinx is a document generator'),
    ]



"""Sphinx core events.

Gracefully adapted from the TextPress system by Armin.
"""

from __future__ import annotations

from collections import defaultdict
from operator import attrgetter
from typing import TYPE_CHECKING, NamedTuple, overload

from sphinx.errors import ExtensionError, SphinxError
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.inspect import safe_getattr

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence, Set
    from pathlib import Path
    from typing import Any, Literal

    from docutils import nodes

    from sphinx import addnodes
    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.config import Config
    from sphinx.domains import Domain
    from sphinx.environment import BuildEnvironment
    from sphinx.ext.autodoc import _AutodocProcessDocstringListener
    from sphinx.ext.todo import todo_node


logger = logging.getLogger(__name__)


class EventListener(NamedTuple):
    id: int
    handler: Callable[..., Any]
    priority: int


# List of all known core events. Maps name to arguments description.
core_events = {
    'config-inited': 'config',
    'builder-inited': '',
    'env-get-outdated': 'env, added, changed, removed',
    'env-before-read-docs': 'env, docnames',
    'env-purge-doc': 'env, docname',
    'source-read': 'docname, source text',
    'include-read': 'relative path, parent docname, source text',
    'doctree-read': 'the doctree before being pickled',
    'env-merge-info': 'env, read docnames, other env instance',
    'env-updated': 'env',
    'env-get-updated': 'env',
    'env-check-consistency': 'env',
    'write-started': 'builder',
    'doctree-resolved': 'doctree, docname',
    'missing-reference': 'env, node, contnode',
    'warn-missing-reference': 'domain, node',
    'build-finished': 'exception',
}


class EventManager:
    """Event manager for Sphinx."""

    def __init__(self, app: Sphinx) -> None:
        self.app = app
        self.events = core_events.copy()
        self.listeners: dict[str, list[EventListener]] = defaultdict(list)
        self.next_listener_id = 0

    def add(self, name: str) -> None:
        """Register a custom Sphinx event."""
        if name in self.events:
            raise ExtensionError(__('Event %r already present') % name)
        self.events[name] = ''

    # ---- Core events -------------------------------------------------------

    @overload
    def connect(
        self,
        name: Literal['config-inited'],
        callback: Callable[[Sphinx, Config], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['builder-inited'],
        callback: Callable[[Sphinx], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-get-outdated'],
        callback: Callable[
            [Sphinx, BuildEnvironment, Set[str], Set[str], Set[str]], Sequence[str]
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-before-read-docs'],
        callback: Callable[[Sphinx, BuildEnvironment, list[str]], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-purge-doc'],
        callback: Callable[[Sphinx, BuildEnvironment, str], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['source-read'],
        callback: Callable[[Sphinx, str, list[str]], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['include-read'],
        callback: Callable[[Sphinx, Path, str, list[str]], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['doctree-read'],
        callback: Callable[[Sphinx, nodes.document], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-merge-info'],
        callback: Callable[
            [Sphinx, BuildEnvironment, Set[str], BuildEnvironment], None
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-updated'],
        callback: Callable[[Sphinx, BuildEnvironment], str],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-get-updated'],
        callback: Callable[[Sphinx, BuildEnvironment], Iterable[str]],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['env-check-consistency'],
        callback: Callable[[Sphinx, BuildEnvironment], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['write-started'],
        callback: Callable[[Sphinx, Builder], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['doctree-resolved'],
        callback: Callable[[Sphinx, nodes.document, str], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['missing-reference'],
        callback: Callable[
            [Sphinx, BuildEnvironment, addnodes.pending_xref, nodes.TextElement],
            nodes.reference | None,
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['warn-missing-reference'],
        callback: Callable[[Sphinx, Domain, addnodes.pending_xref], bool | None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['build-finished'],
        callback: Callable[[Sphinx, Exception | None], None],
        priority: int,
    ) -> int: ...

    # ---- Events from builtin builders --------------------------------------

    @overload
    def connect(
        self,
        name: Literal['html-collect-pages'],
        callback: Callable[[Sphinx], Iterable[tuple[str, dict[str, Any], str]]],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['html-page-context'],
        callback: Callable[
            [Sphinx, str, str, dict[str, Any], nodes.document], str | None
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['linkcheck-process-uri'],
        callback: Callable[[Sphinx, str], str | None],
        priority: int,
    ) -> int: ...

    # ---- Events from builtin extensions-- ----------------------------------

    @overload
    def connect(
        self,
        name: Literal['object-description-transform'],
        callback: Callable[[Sphinx, str, str, addnodes.desc_content], None],
        priority: int,
    ) -> int: ...

    # ---- Events from first-party extensions --------------------------------

    @overload
    def connect(
        self,
        name: Literal['autodoc-process-docstring'],
        callback: _AutodocProcessDocstringListener,
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['autodoc-before-process-signature'],
        callback: Callable[[Sphinx, Any, bool], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['autodoc-process-signature'],
        callback: Callable[
            [
                Sphinx,
                Literal[
                    'module', 'class', 'exception', 'function', 'method', 'attribute'
                ],
                str,
                Any,
                dict[str, bool],
                str | None,
                str | None,
            ],
            tuple[str | None, str | None] | None,
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['autodoc-process-bases'],
        callback: Callable[[Sphinx, str, Any, dict[str, bool], list[str]], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['autodoc-skip-member'],
        callback: Callable[
            [
                Sphinx,
                Literal[
                    'module', 'class', 'exception', 'function', 'method', 'attribute'
                ],
                str,
                Any,
                bool,
                dict[str, bool],
            ],
            bool,
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['todo-defined'],
        callback: Callable[[Sphinx, todo_node], None],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['viewcode-find-source'],
        callback: Callable[
            [Sphinx, str],
            tuple[str, dict[str, tuple[Literal['class', 'def', 'other'], int, int]]],
        ],
        priority: int,
    ) -> int: ...

    @overload
    def connect(
        self,
        name: Literal['viewcode-follow-imported'],
        callback: Callable[[Sphinx, str, str], str | None],
        priority: int,
    ) -> int: ...

    # ---- Catch-all ---------------------------------------------------------

    @overload
    def connect(
        self,
        name: str,
        callback: Callable[..., Any],
        priority: int,
    ) -> int: ...

    def connect(self, name: str, callback: Callable[..., Any], priority: int) -> int:
        """Connect a handler to specific event."""
        if name not in self.events:
            raise ExtensionError(__('Unknown event name: %s') % name)

        listener_id = self.next_listener_id
        self.next_listener_id += 1
        self.listeners[name].append(EventListener(listener_id, callback, priority))
        return listener_id

    def disconnect(self, listener_id: int) -> None:
        """Disconnect a handler."""
        for listeners in self.listeners.values():
            for listener in listeners.copy():
                if listener.id == listener_id:
                    listeners.remove(listener)

    def emit(
        self,
        name: str,
        *args: Any,
        allowed_exceptions: tuple[type[Exception], ...] = (),
    ) -> list[Any]:
        """Emit a Sphinx event."""
        # not every object likes to be repr()'d (think
        # random stuff coming via autodoc)
        try:
            repr_args = repr(args)
        except Exception:
            pass
        else:
            logger.debug('[app] emitting event: %r%s', name, repr_args)

        results = []
        listeners = sorted(self.listeners[name], key=attrgetter('priority'))
        for listener in listeners:
            try:
                results.append(listener.handler(self.app, *args))
            except allowed_exceptions:
                # pass through the errors specified as *allowed_exceptions*
                raise
            except SphinxError:
                raise
            except Exception as exc:
                if self.app.pdb:
                    # Just pass through the error, so that it can be debugged.
                    raise
                modname = safe_getattr(listener.handler, '__module__', None)
                raise ExtensionError(
                    __('Handler %r for event %r threw an exception')
                    % (listener.handler, name),
                    exc,
                    modname=modname,
                ) from exc
        return results

    def emit_firstresult(
        self,
        name: str,
        *args: Any,
        allowed_exceptions: tuple[type[Exception], ...] = (),
    ) -> Any:
        """Emit a Sphinx event and returns first result.

        This returns the result of the first handler that doesn't return ``None``.
        """
        for result in self.emit(name, *args, allowed_exceptions=allowed_exceptions):
            if result is not None:
                return result
        return None



"""Tests sphinx.util.template functions."""

from __future__ import annotations

from sphinx.util.template import ReSTRenderer


def test_ReSTRenderer_escape() -> None:
    r = ReSTRenderer()
    template = '{{ "*hello*" | e }}'
    assert r.render_string(template, {}) == r'\*hello\*'


def test_ReSTRenderer_heading() -> None:
    r = ReSTRenderer()

    template = '{{ "hello" | heading }}'
    assert r.render_string(template, {}) == 'hello\n====='

    template = '{{ "hello" | heading(1) }}'
    assert r.render_string(template, {}) == 'hello\n====='

    template = '{{ "русский язык" | heading(2) }}'
    assert r.render_string(template, {}) == 'русский язык\n------------'

    # language: ja
    r.env.language = 'ja'  # type: ignore[attr-defined]
    template = '{{ "русский язык" | heading }}'
    assert r.render_string(template, {}) == 'русский язык\n======================='



project = 'Sphinx'
version = '1.4.4'

html_static_path = ['static', 'subdir']
html_extra_path = ['extra', 'subdir']
html_css_files = [
    'css/style.css',
    (
        'https://example.com/custom.css',
        {'title': 'title', 'media': 'print', 'priority': 400},
    ),
]
html_js_files = [
    'js/custom.js',
    (
        'https://example.com/script.js',
        {'async': 'async', 'priority': 400},
    ),
]
exclude_patterns = ['**/_build', '**/.htpasswd']



"""Test pycode.ast"""

from __future__ import annotations

import ast

import pytest

from sphinx.pycode.ast import unparse as ast_unparse


@pytest.mark.parametrize(
    ('source', 'expected'),
    [
        ('a + b', 'a + b'),                        # Add
        ('a and b', 'a and b'),                    # And
        ('os.path', 'os.path'),                    # Attribute
        ('1 * 2', '1 * 2'),                        # BinOp
        ('a & b', 'a & b'),                        # BitAnd
        ('a | b', 'a | b'),                        # BitOr
        ('a ^ b', 'a ^ b'),                        # BitXor
        ('a and b and c', 'a and b and c'),        # BoolOp
        ("b'bytes'", "b'bytes'"),                  # Bytes
        ('object()', 'object()'),                  # Call
        ('1234', '1234'),                          # Constant, Num
        ("{'key1': 'value1', 'key2': 'value2'}",
         "{'key1': 'value1', 'key2': 'value2'}"),  # Dict
        ('a / b', 'a / b'),                        # Div
        ('...', '...'),                            # Ellipsis
        ('a // b', 'a // b'),                      # FloorDiv
        ('Tuple[int, int]', 'Tuple[int, int]'),    # Index, Subscript
        ('~1', '~1'),                              # Invert
        ('lambda x, y: x + y',
         'lambda x, y: ...'),                      # Lambda
        ('[1, 2, 3]', '[1, 2, 3]'),                # List
        ('a << b', 'a << b'),                      # LShift
        ('a @ b', 'a @ b'),                        # MatMult
        ('a % b', 'a % b'),                        # Mod
        ('a * b', 'a * b'),                        # Mult
        ('sys', 'sys'),                            # Name, NameConstant
        ('not a', 'not a'),                        # Not
        ('a or b', 'a or b'),                      # Or
        ('a**b', 'a**b'),                          # Pow
        ('a >> b', 'a >> b'),                      # RShift
        ('{1, 2, 3}', '{1, 2, 3}'),                # Set
        ('a - b', 'a - b'),                        # Sub
        ("'str'", "'str'"),                        # Str
        ('+a', '+a'),                              # UAdd
        ('-1', '-1'),                              # UnaryOp
        ('-a', '-a'),                              # USub
        ('(1, 2, 3)', '(1, 2, 3)'),                # Tuple
        ('()', '()'),                              # Tuple (empty)
        ('(1,)', '(1,)'),                          # Tuple (single item)
        ('lambda x=0, /, y=1, *args, z, **kwargs: x + y + z',
         'lambda x=0, /, y=1, *args, z, **kwargs: ...'),  # posonlyargs
        ('0x1234', '0x1234'),                      # Constant
        ('1_000_000', '1_000_000'),                # Constant
        ('Tuple[:,:]', 'Tuple[:, :]'),             # Index, Subscript, 2x Slice
        ('Tuple[1:2]', 'Tuple[1:2]'),              # Index, Subscript, Slice(no-step)
        ('Tuple[1:2:3]', 'Tuple[1:2:3]'),          # Index, Subscript, Slice
        ('x[:, np.newaxis, :, :]',
         'x[:, np.newaxis, :, :]'),                # Index, Subscript, numpy extended syntax
        ('y[:, 1:3][np.array([0, 2, 4]), :]',
         'y[:, 1:3][np.array([0, 2, 4]), :]'),     # Index, 2x Subscript, numpy extended syntax
    ],
)  # fmt: skip
def test_unparse(source, expected):
    module = ast.parse(source)
    assert ast_unparse(module.body[0].value, source) == expected


def test_unparse_None() -> None:
    assert ast_unparse(None) is None



"""Test the ChangesBuilder class."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('changes', testroot='changes')
def test_build(app: SphinxTestApp) -> None:
    app.build()

    # TODO: Use better checking of html content
    htmltext = (app.outdir / 'changes.html').read_text(encoding='utf8')
    assert 'Added in version 0.6: Some funny stuff.' in htmltext
    assert 'Changed in version 0.6: Even more funny stuff.' in htmltext
    assert 'Deprecated since version 0.6: Boring stuff.' in htmltext

    path_html = (
        '<b>Path</b>: <i>deprecated:</i> Deprecated since version 0.6:'
        ' So, that was a bad idea it turns out.'
    )
    assert path_html in htmltext

    malloc_html = (
        '<b>void *Test_Malloc(size_t n)</b>: <i>changed:</i> Changed in version 0.6:'
        ' Can now be replaced with a different allocator.</a>'
    )
    assert malloc_html in htmltext


@pytest.mark.sphinx(
    'changes',
    testroot='changes',
    srcdir='changes-none',
    confoverrides={'version': '0.7', 'release': '0.7b1'},
)
def test_no_changes(app: SphinxTestApp) -> None:
    app.build()

    assert 'no changes in version 0.7.' in app.status.getvalue()
    assert not (app.outdir / 'changes.html').exists()



# Literally included file using Python highlighting

foo = 'Including Unicode characters: üöä'


class Foo:
    pass


class Bar:
    def baz():
        pass


# comment after Bar class definition
def bar(): pass


def block_start_with_comment():
    # Comment
    return 1


def block_start_with_blank():

    return 1


class Qux:
	def quux(self):
		pass



"""Test the EventManager class."""

from __future__ import annotations

from types import SimpleNamespace

import pytest

from sphinx.errors import ExtensionError
from sphinx.events import EventManager


def test_event_priority() -> None:
    result = []
    app = object()  # pass a dummy object as an app
    events = EventManager(app)  # type: ignore[arg-type]
    events.connect('builder-inited', lambda app: result.append(1), priority=500)
    events.connect('builder-inited', lambda app: result.append(2), priority=500)
    # earlier
    events.connect('builder-inited', lambda app: result.append(3), priority=200)
    # later
    events.connect('builder-inited', lambda app: result.append(4), priority=700)
    events.connect('builder-inited', lambda app: result.append(5), priority=500)

    events.emit('builder-inited')
    assert result == [3, 1, 2, 5, 4]


def test_event_allowed_exceptions() -> None:
    def raise_error(app):
        raise RuntimeError

    app = SimpleNamespace(pdb=False)  # pass a dummy object as an app
    events = EventManager(app)  # type: ignore[arg-type]
    events.connect('builder-inited', raise_error, priority=500)

    # all errors are converted to ExtensionError
    with pytest.raises(ExtensionError):
        events.emit('builder-inited')

    # Allow RuntimeError (pass-through)
    with pytest.raises(RuntimeError):
        events.emit('builder-inited', allowed_exceptions=(RuntimeError,))


def test_event_pdb() -> None:
    def raise_error(app):
        raise RuntimeError

    app = SimpleNamespace(pdb=True)  # pass a dummy object as an app
    events = EventManager(app)  # type: ignore[arg-type]
    events.connect('builder-inited', raise_error, priority=500)

    # errors aren't converted
    with pytest.raises(RuntimeError):
        events.emit('builder-inited')

    # Allow RuntimeError (pass-through)
    with pytest.raises(RuntimeError):
        events.emit('builder-inited', allowed_exceptions=(RuntimeError,))



from pathlib import Path

import mod_resource
import mod_something

if __name__ == '__main__':
    print(f'Hello, world! -> something returns: {mod_something.something()}')

    res_path = Path(mod_resource.__file__).parent / 'resource.txt'
    text = res_path.read_text(encoding='utf-8')
    print(f'From mod_resource:resource.txt -> {text}')



import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().resolve()))

extensions = ['sphinx.ext.autosummary']
autosummary_generate = True
autosummary_filename_map = {
    'autosummary_dummy_module': 'module_mangled',
    'autosummary_dummy_module.bar': 'bar',
}



"""Tests sphinx.util.fileutil functions."""

from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.jinja2glue import BuiltinTemplateLoader
from sphinx.util.fileutil import _template_basename, copy_asset, copy_asset_file

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


class DummyTemplateLoader(BuiltinTemplateLoader):
    def __init__(self) -> None:
        super().__init__()
        builder = mock.Mock()
        builder.config.templates_path = []
        builder._translator = None
        self.init(builder)


def test_copy_asset_file(tmp_path):
    renderer = DummyTemplateLoader()

    # copy normal file
    src = tmp_path / 'asset.txt'
    src.write_text('# test data', encoding='utf8')
    dest = tmp_path / 'output.txt'

    copy_asset_file(src, dest)
    assert dest.exists()
    assert src.read_text(encoding='utf8') == dest.read_text(encoding='utf8')

    # copy template file
    src = tmp_path / 'asset.txt.jinja'
    src.write_text('# {{var1}} data', encoding='utf8')
    dest = tmp_path / 'output.txt.jinja'

    copy_asset_file(str(src), str(dest), {'var1': 'template'}, renderer)
    assert not dest.exists()
    assert (tmp_path / 'output.txt').exists()
    assert (tmp_path / 'output.txt').read_text(encoding='utf8') == '# template data'

    # copy template file to subdir
    src = tmp_path / 'asset.txt.jinja'
    src.write_text('# {{var1}} data', encoding='utf8')
    subdir1 = tmp_path / 'subdir'
    subdir1.mkdir(parents=True, exist_ok=True)

    copy_asset_file(src, subdir1, context={'var1': 'template'}, renderer=renderer)
    assert (subdir1 / 'asset.txt').exists()
    assert (subdir1 / 'asset.txt').read_text(encoding='utf8') == '# template data'

    # copy template file without context
    src = tmp_path / 'asset.txt.jinja'
    subdir2 = tmp_path / 'subdir2'
    subdir2.mkdir(parents=True, exist_ok=True)

    copy_asset_file(src, subdir2)
    assert not (subdir2 / 'asset.txt').exists()
    assert (subdir2 / 'asset.txt.jinja').exists()
    assert (subdir2 / 'asset.txt.jinja').read_text(encoding='utf8') == '# {{var1}} data'


def test_copy_asset(tmp_path):
    renderer = DummyTemplateLoader()

    # prepare source files
    source = tmp_path / 'source'
    source.mkdir(parents=True, exist_ok=True)
    (source / 'index.rst').write_text('index.rst', encoding='utf8')
    (source / 'foo.rst.jinja').write_text('{{var1}}.rst', encoding='utf8')
    (source / '_static').mkdir(parents=True, exist_ok=True)
    (source / '_static' / 'basic.css').write_text('basic.css', encoding='utf8')
    (source / '_templates').mkdir(parents=True, exist_ok=True)
    (source / '_templates' / 'layout.html').write_text('layout.html', encoding='utf8')
    (source / '_templates' / 'sidebar.html.jinja').write_text(
        'sidebar: {{var2}}', encoding='utf8'
    )
    sibling = tmp_path / 'sibling'
    sibling.mkdir(parents=True, exist_ok=True)
    sibling.joinpath('spam').mkdir(parents=True, exist_ok=True)
    sibling.joinpath('spam', 'ham').touch()

    # copy a single file
    assert not (tmp_path / 'test1').exists()
    copy_asset(source / 'index.rst', tmp_path / 'test1')
    assert (tmp_path / 'test1').exists()
    assert (tmp_path / 'test1/index.rst').exists()

    # copy directories
    destdir = tmp_path / 'test2'
    copy_asset(
        source,
        destdir,
        context={'var1': 'bar', 'var2': 'baz'},
        renderer=renderer,
    )
    assert (destdir / 'index.rst').exists()
    assert (destdir / 'foo.rst').exists()
    assert (destdir / 'foo.rst').read_text(encoding='utf8') == 'bar.rst'
    assert (destdir / '_static' / 'basic.css').exists()
    assert (destdir / '_templates' / 'layout.html').exists()
    assert (destdir / '_templates' / 'sidebar.html').exists()
    sidebar = (destdir / '_templates' / 'sidebar.html').read_text(encoding='utf8')
    assert sidebar == 'sidebar: baz'

    # copy with exclusion
    def excluded(path):
        return 'sidebar.html' in path or 'basic.css' in path

    destdir = tmp_path / 'test3'
    copy_asset(
        source,
        destdir,
        excluded,
        context={'var1': 'bar', 'var2': 'baz'},
        renderer=renderer,
    )
    assert (destdir / 'index.rst').exists()
    assert (destdir / 'foo.rst').exists()
    assert not (destdir / '_static' / 'basic.css').exists()
    assert (destdir / '_templates' / 'layout.html').exists()
    assert not (destdir / '_templates' / 'sidebar.html').exists()

    # copy sibling
    assert not (tmp_path / 'test4').exists()
    copy_asset(source / '../sibling', tmp_path / 'test4')
    assert (tmp_path / 'test4').is_dir()
    assert (tmp_path / 'test4' / 'spam').is_dir()
    assert (tmp_path / 'test4' / 'spam' / 'ham').is_file()
    assert (tmp_path / 'test4' / 'spam' / 'ham').read_bytes() == b''


@pytest.mark.sphinx('html', testroot='html_assets')
def test_copy_asset_template(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    expected_msg = r'^Writing evaluated template result to [^\n]*\bAPI.html$'
    output = strip_escape_sequences(app.status.getvalue())
    assert re.findall(expected_msg, output, flags=re.MULTILINE)


@pytest.mark.sphinx('html', testroot='util-copyasset_overwrite')
def test_copy_asset_overwrite(app: SphinxTestApp) -> None:
    app.build()
    src = app.srcdir / 'myext_static' / 'custom-styles.css'
    dst = app.outdir / '_static' / 'custom-styles.css'
    assert strip_escape_sequences(app.warning.getvalue()) == (
        f'WARNING: Aborted attempted copy from {src} to {dst} '
        '(the destination path has existing data). '
        '[misc.copy_overwrite]\n'
    )


def test_template_basename() -> None:
    assert _template_basename(Path('asset.txt')) is None
    assert _template_basename(Path('asset.txt.jinja')) == Path('asset.txt')
    assert _template_basename(Path('sidebar.html.jinja')) == Path('sidebar.html')


def test_legacy_template_basename() -> None:
    assert _template_basename(Path('asset.txt_t')) == Path('asset.txt')



"""Tests for ``record_dependencies``."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.util._pathlib import _StrPath

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='environment-record-dependencies')
def test_record_dependencies_cleared(app: SphinxTestApp) -> None:
    app.builder.read()
    assert 'index' not in app.env.dependencies
    assert app.env.dependencies['api'] == {_StrPath('example_module.py')}



"""Test autosummary for import cycles."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.ext.autosummary import autosummary_table
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('dummy', testroot='ext-autosummary-import_cycle')
@pytest.mark.usefixtures('rollback_sysmodules')
def test_autosummary_import_cycle(app):
    app.build()

    doctree = app.env.get_doctree('index')
    app.env.apply_post_transforms(doctree, 'index')

    assert len(list(doctree.findall(nodes.reference))) == 1

    assert_node(
        doctree,
        (
            addnodes.index,  # [0]
            nodes.target,  # [1]
            nodes.paragraph,  # [2]
            addnodes.tabular_col_spec,  # [3]
            [
                autosummary_table,
                nodes.table,
                nodes.tgroup,
                (
                    nodes.colspec,  # [4][0][0][0]
                    nodes.colspec,  # [4][0][0][1]
                    [nodes.tbody, nodes.row],  # [4][0][0][2][1]
                ),
            ],
            addnodes.index,  # [5]
            addnodes.desc,  # [6]
        ),
    )
    assert_node(
        doctree[4][0][0][2][0],
        ([nodes.entry, nodes.paragraph, (nodes.reference, nodes.Text)], nodes.entry),
    )
    assert_node(
        doctree[4][0][0][2][0][0][0][0],
        nodes.reference,
        refid='spam.eggs.Ham',
        reftitle='spam.eggs.Ham',
    )

    expected = (
        'Summarised items should not include the current module. '
        "Replace 'spam.eggs.Ham' with 'Ham'."
    )
    assert expected in app.warning.getvalue()


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-module_prefix',
    copy_test_root=True,
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_autosummary_generate_prefixes(app: SphinxTestApp) -> None:
    app.build()
    warnings = app.warning.getvalue()
    assert 'Summarised items should not include the current module.' not in warnings
    assert warnings == ''



from functools import partial


def func():
    pass


async def coroutinefunc():
    pass


async def asyncgenerator():  # NoQA: RUF029
    yield


partial_func = partial(func)
partial_coroutinefunc = partial(coroutinefunc)

builtin_func = print
partial_builtin_func = partial(print)


def slice_arg_func(arg: 'float64[:, :]'):  # NoQA: F821
    pass
